{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "c8bcf327",
   "metadata": {},
   "outputs": [],
   "source": [
    "import collections\n",
    "import copy\n",
    "import itertools\n",
    "import json\n",
    "import math\n",
    "import os\n",
    "import random\n",
    "import sqlite3\n",
    "from ogb.nodeproppred import DglNodePropPredDataset\n",
    "import cProfile\n",
    "import dgl\n",
    "import torch\n",
    "from tqdm import tqdm\n",
    "from torch.sparse import *\n",
    "from tabulate import tabulate\n",
    "import numpy as np\n",
    "\n",
    "os.environ['DGLBACKEND'] = 'pytorch'"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "f5ecdcab",
   "metadata": {},
   "outputs": [],
   "source": [
    "class Unpacker():\n",
    "    \"\"\"input of quotiented_simplices is a dictionary with keys = (dimension, simplex_id, globe_id)\n",
    "    and values = [v_0,v_1,...,v_n]. Node mapping is also a dictionary with keys = nodes\n",
    "    and values = a list of pre-images. Such a list could be a singleton, but never empty\n",
    "    because the quotient map is a function in the math sense. The other inputs concern the graph downstairs (range graph)\n",
    "    and the graph upstairs (domain graph).\n",
    "    \n",
    "    This function pushes simplices from downstairs to upstairs, and does this by looking at degenerate simplice downstairs,\n",
    "    then finds possible preimages for them upstairs, and does so inductively. The highest dimension for which we need\n",
    "    to look for a simplex is dictated by the graph upstairs, and is decided by self.local_max_dim.\n",
    "    \n",
    "    Finally, the simplex_max_index takes the index from the previous section finder, and increments it for simplices\n",
    "    that are found later.\"\"\"\n",
    "    \n",
    "    def __init__(self, domain_graph,node_mapping,range_graph_edges,\n",
    "                 deg_edges,quotiented_simplices,simplex_max_index,globe_number,domain_graph_edges):\n",
    "        \n",
    "        assert isinstance(domain_graph, dgl.DGLHeteroGraph), \\\n",
    "        'Keyword argument \\\"domain_graph\\\" must be a dgl.DGLHeteroGraph.'\n",
    "        assert isinstance(quotiented_simplices, dict), \\\n",
    "        'Keyword argument \\\"edges_list\\\" must be a dictionary.'\n",
    "        assert isinstance(node_mapping, dict), \\\n",
    "        'Keyword argument \\\"node_mapping\\\" must be a dictionary.'\n",
    "        assert isinstance(deg_edges, list), \\\n",
    "        'Keyword argument \\\"deg_edges\\\" must be a dictionary.'\n",
    "\n",
    "        self.simplex_max_index     = simplex_max_index\n",
    "        self.domain_graph          = domain_graph\n",
    "        self.node_mapping          = node_mapping\n",
    "        self.deg_edges             = deg_edges\n",
    "        self.range_graph_edges     = range_graph_edges\n",
    "        self.quotiented_simplices  = quotiented_simplices\n",
    "        self.lifted_simplices      = dict()\n",
    "        self.globe_number          = globe_number\n",
    "        number_of_nodes            = len(torch.unique(torch.cat(self.domain_graph.edges())))\n",
    "        in_degrees                 = self.domain_graph.in_degrees()\n",
    "        out_degrees                = self.domain_graph.out_degrees()\n",
    "        possible_max1              = min(int(torch.max(in_degrees)), number_of_nodes-1)\n",
    "        possible_max2              = min(int(torch.max(out_degrees)), number_of_nodes-1)\n",
    "        self.local_max_dim         = max(possible_max1,possible_max2)\n",
    "        self.domain_graph_edges    = domain_graph_edges\n",
    "        self.preimages_collection  = list()\n",
    "        self.injective_nodes       = list()\n",
    "        self.non_injective_nodes   = list()\n",
    "        self.truquotiented_simpl   = copy.deepcopy(quotiented_simplices)\n",
    "        self.new_simplices         = list()\n",
    "        self.all_range_g_edges     = self.range_graph_edges + self.deg_edges\n",
    "        \n",
    "        # Seperate the nodes that are mapped injectively\n",
    "        for node, preimages in self.node_mapping.items():\n",
    "            if len(preimages) == 1:\n",
    "                self.injective_nodes = self.injective_nodes + [node]\n",
    "            else:\n",
    "                self.non_injective_nodes = self.non_injective_nodes + [node]\n",
    "        \n",
    "        index = 0\n",
    "        _nodes = set(self.non_injective_nodes)\n",
    "        for key, quotiented_simplex in self.quotiented_simplices.items():\n",
    "            if key[0] > 1:\n",
    "                common_elements = _nodes.intersection(set(quotiented_simplex))\n",
    "                if len(common_elements) == 0:\n",
    "                    # If the simplex downstairs is the same as the simplex upstairs, then add it to the dictionary\n",
    "                    key_lifted = (len(quotiented_simplex)-1, self.simplex_max_index+index, self.globe_number)\n",
    "                    self.lifted_simplices.update({key_lifted:quotiented_simplex})\n",
    "                    #T he simplex found is not a degenerate simplex upstairs, so we don't need to find its sections.\n",
    "                    del self.truquotiented_simpl[key]\n",
    "                    index = index + 1\n",
    "                else:\n",
    "                    # Work out the simplices upstairs that have images of the same dimension downstairs\n",
    "                    degeneracies = self.generate_youngs_list(quotiented_simplex)\n",
    "                    for degeneracy in degeneracies:\n",
    "                        if self.simplex_verifier(degeneracy):\n",
    "                            key_lift = (len(degeneracy)-1, self.simplex_max_index+index, self.globe_number)\n",
    "                            self.lifted_simplices.update({key_lift:degeneracy})\n",
    "                            index = index + 1\n",
    "        self.simplex_max_index = self.simplex_max_index + index\n",
    "        self.new_simplices = list(self.truquotiented_simpl.values()) + self.range_graph_edges\n",
    "        \n",
    "    def inductive_connecting(self):          \n",
    "        new_zero_skeleta = self.new_simplices\n",
    "        self.images = []\n",
    "        for src in self.non_injective_nodes:\n",
    "            for zero_skel in new_zero_skeleta:\n",
    "                if len(zero_skel) > self.local_max_dim:\n",
    "                    continue\n",
    "                if zero_skel.count(src) >= len(self.node_mapping[src]):\n",
    "                    continue\n",
    "                if src not in zero_skel:\n",
    "                    continue\n",
    "                edge_present_query1 = True\n",
    "                edge_present_query2 = True\n",
    "                for dst in zero_skel:\n",
    "                    if edge_present_query1:\n",
    "                        if [src, dst] in self.all_range_g_edges:\n",
    "                            index_to_insert = zero_skel.index(src)\n",
    "                            if zero_skel[:index_to_insert]+[src]+zero_skel[index_to_insert:] not in self.images:\n",
    "                                self.images.append(zero_skel[:index_to_insert]+[src]+zero_skel[index_to_insert:])\n",
    "                        if [dst, src] in self.all_range_g_edges:\n",
    "                            index_to_insert = zero_skel.index(dst)\n",
    "                            if zero_skel[:index_to_insert-1]+[src]+zero_skel[index_to_insert-1:] not in self.images:\n",
    "                                self.images.append(zero_skel[:index_to_insert-1]+[src]+zero_skel[index_to_insert-1:])\n",
    "       \n",
    "        if len(self.images) == 0:\n",
    "            # No more images are found, so we can stop the induction process\n",
    "            return\n",
    "        \n",
    "        else:\n",
    "            self.preimages_collection = []\n",
    "            self.preimages_collection_id = dict()\n",
    "            for image in self.images:\n",
    "                preimages = self.generate_youngs_list(image)\n",
    "                for preimage in preimages:\n",
    "                    self.preimages_collection_id.update({id(preimage):image})\n",
    "                self.preimages_collection = self.preimages_collection + preimages\n",
    "            self.add_simplices_to_dict()\n",
    "            self.new_simplices = [list(image) for image in self.simplices_to_extend]\n",
    "            self.inductive_connecting()\n",
    "    \n",
    "    def generate_youngs_list(self, simplex):\n",
    "        \"\"\"Takes as input the image of a quotiented simplex spits out all possible simplices in the pre-image\"\"\"\n",
    "        \n",
    "        indices = [range(len(self.node_mapping[key])) for key in simplex]\n",
    "        combinations = itertools.product(*indices)\n",
    "        result_lists = []\n",
    "        for combo in combinations:\n",
    "            alternative_list = [self.node_mapping[simplex[0]][combo[0]]]\n",
    "            for i in range(1, len(simplex)):\n",
    "                to_add = True\n",
    "                p_node = self.node_mapping[simplex[i-1]][combo[i-1]]\n",
    "                c_node = self.node_mapping[simplex[i]][combo[i]]\n",
    "                if (p_node,c_node) in self.domain_graph_edges:\n",
    "                    if c_node not in alternative_list: \n",
    "                        alternative_list.append(self.node_mapping[simplex[i]][combo[i]])\n",
    "                    else:\n",
    "                        to_add = False\n",
    "                        break\n",
    "                else:\n",
    "                    to_add = False\n",
    "                    break\n",
    "            if to_add:\n",
    "                if alternative_list not in result_lists:\n",
    "                    result_lists.append(alternative_list)\n",
    "        return result_lists\n",
    "\n",
    "    def simplex_verifier(self, simplex):\n",
    "        \"\"\"This function takes input a k-tuple [v_1,v_2,...,v_k] for k>1 and a graph G\n",
    "        and verifies if it a simplex of G. This is done by simply checking if there is \n",
    "        an edge between v_i and v_j for i<j\"\"\"\n",
    "    \n",
    "        is_simplex = False\n",
    "        simplex_reduced = copy.deepcopy(simplex)\n",
    "        \n",
    "        for i in simplex:\n",
    "            simplex_reduced.pop(0)\n",
    "            for j in simplex_reduced:\n",
    "                if (i, j) in self.domain_graph_edges:\n",
    "                    continue\n",
    "                else:\n",
    "                    # The input is not a simplex, so we can stop our check.\n",
    "                    return is_simplex\n",
    "        return True\n",
    "    \n",
    "    def add_simplices_to_dict(self):\n",
    "        index = 0\n",
    "        self.simplices_to_extend = set()\n",
    "\n",
    "        for preimage in self.preimages_collection:\n",
    "            if self.simplex_verifier(preimage):\n",
    "                key = (len(preimage) - 1, self.simplex_max_index + index, self.globe_number)\n",
    "                self.lifted_simplices.update({key: preimage})\n",
    "                index += 1\n",
    "                temp_dict = {tuple(self.preimages_collection_id[id(preimage)])}\n",
    "                self.simplices_to_extend = self.simplices_to_extend.union(temp_dict)\n",
    "        self.simplex_max_index = self.simplex_max_index + index \n",
    "    \n",
    "    def simplex_breaker(self,simplex):\n",
    "        \"\"\"This function takes a simplex and returns all the edges within it. I abondoned the plan to use it.\"\"\"\n",
    "        edges = list()\n",
    "        for i in range(len(simplex)):\n",
    "            for j in range(i + 1, len(simplex)):\n",
    "                x = lst[i]\n",
    "                y = lst[j]\n",
    "                edges = edges + [[x,y]]\n",
    "        return edges"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "bf854178",
   "metadata": {},
   "outputs": [],
   "source": [
    "def edges_to_identify(graph):\n",
    "    \"\"\"Constructing the Hadamard product here to find the list of edges to contract.\n",
    "    All the edges that are parts of a 2-simplex are identified in this function\n",
    "    as indices of the Hadamard product. Moreover, other edges that are not part of the Hadamard\n",
    "    product, but are part of the transitive closure of existing edges are also identified.\n",
    "    These are the edges that form a complete 2-simplex.\"\"\"\n",
    "\n",
    "    assert isinstance(graph, dgl.DGLHeteroGraph), \\\n",
    "        'Keyword argument \\\"graph\\\" of create_hadamard must be a dgl.DGLHeteroGraph.'\n",
    "    \n",
    "    # First, we remove all diag entries from adj matrix A by removing all self-loops\n",
    "    loopless = dgl.transforms.RemoveSelfLoop()\n",
    "    graph = loopless(graph)\n",
    "    \n",
    "    # Then we remove all diagonals from A^2 and convert the matrix to one with binary entries\n",
    "    adj_squared = torch.sparse.mm(graph.adj_external(),graph.adj_external())\n",
    "    diagonal_mask = (adj_squared._indices()[0] == adj_squared._indices()[1])\n",
    "    off_diagonal_mask = ~diagonal_mask\n",
    "    adj_squared._values()[off_diagonal_mask] = 1.0\n",
    "    new_indices = adj_squared._indices()[:, off_diagonal_mask]\n",
    "    new_values = adj_squared._values()[off_diagonal_mask]\n",
    "    new_size = adj_squared.size()\n",
    "    squared_no_diag_binary = torch.sparse_coo_tensor(indices=new_indices, \n",
    "                                                    values=new_values, size=new_size)\n",
    "    # The Hadamard product is sparse, but keeps track of entries that are zero.\n",
    "    false_hadamard_product = graph.adj_external() * squared_no_diag_binary\n",
    "    \n",
    "    # We, therefore need to remove those entries.\n",
    "    false_hadamard_product = false_hadamard_product.coalesce()\n",
    "    non_zero_mask = false_hadamard_product._values().nonzero().squeeze()\n",
    "    non_zero_values = false_hadamard_product._values()[non_zero_mask]\n",
    "    non_zero_indices = false_hadamard_product.indices()[:, non_zero_mask]\n",
    "    if non_zero_indices.dim() == 1:\n",
    "        non_zero_indices = non_zero_indices.unsqueeze(0)\n",
    "        non_zero_indices = non_zero_indices.view(2, -1)\n",
    "    hadamard_product = torch.sparse_coo_tensor(indices=non_zero_indices,\n",
    "                                               values=non_zero_values,\n",
    "                                               size=false_hadamard_product.size())\n",
    "    \n",
    "    # The following loop finds all edges that are part of a simplex need to be collapsed.\n",
    "    row_indices, col_indices = hadamard_product._indices()\n",
    "    extra_edges = list()\n",
    "    \n",
    "    for i,j in zip(row_indices,col_indices):\n",
    "        out_nodes = set([int(v) for v in list(graph.successors(i))])\n",
    "        in_nodes = set([int(v) for v in list(graph.predecessors(j))])\n",
    "        # These are the elements in the (reverse?) transitive closure of (i,j).\n",
    "        intersection = set.intersection(out_nodes,in_nodes)\n",
    "        for k in intersection:\n",
    "            extra_edges = extra_edges + [(int(i),int(k))] + [(int(k),int(j))]\n",
    "    return hadamard_product, extra_edges"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "0c32d61b",
   "metadata": {},
   "outputs": [],
   "source": [
    "class graph_towers():\n",
    "    src=list()\n",
    "    dst=list()\n",
    "    empty_graph = dgl.heterograph({('node', 'to', 'node'): (src, dst)})\n",
    "    ratio = 0.0\n",
    "    bottom_level = 0\n",
    "    \n",
    "    assert isinstance(empty_graph, dgl.DGLHeteroGraph), \\\n",
    "        'Keyword argument \\\"graph\\\" of graph_towers\\'s init method must be a dgl.DGLHeteroGraph.'\n",
    "    assert isinstance(ratio, float), \\\n",
    "        'Keyword argument \\\"ratio\\\" of graph_towers\\'s init method must be a float.'\n",
    "    assert ratio<=1 and ratio>=0, \\\n",
    "        'Keyword argument \\\"ratio\\\" of graph_towers\\'s init method must be between 0 and 1.'\n",
    "    assert isinstance(bottom_level, int), \\\n",
    "        'Keyword argument \\\"bottom_level\\\" of graph_towers\\'s init method must be an integer.'\n",
    "\n",
    "    def __init__(self, file_path, ratio, database_name, max_dimension, graph=empty_graph,\n",
    "                 bottom_level=bottom_level):\n",
    "        \n",
    "        self.seed_graph        = graph\n",
    "        self.srcs_and_dsts     = self.seed_graph.edges()   \n",
    "        self.file_path         = file_path        \n",
    "        self.ratio             = ratio\n",
    "        self.updated_graph     = dgl.heterograph({('node', 'to', 'node'): ([], [])})\n",
    "        self.bottom_level      = bottom_level\n",
    "        self.database_name     = database_name\n",
    "        self.maximum_dimension = max_dimension\n",
    "        self.connection        = None\n",
    "        self.cursor            = None\n",
    "        self.number_of_nodes   = len(self.seed_graph.nodes())\n",
    "        self.number_of_edges   = len(self.seed_graph.edges()[0])\n",
    "        self.selected_edges    = None\n",
    "        self.quotient_number   = 0\n",
    "        self.simplex_id        = 0\n",
    "        \n",
    "        # Find list of edges that will be used to create a quotient graph\n",
    "        self.hadamard_product, self.extra_edges  = edges_to_identify(self.seed_graph)\n",
    "        rows, columns = self.hadamard_product._indices()\n",
    "        self.edges_to_collapse_as_pairs = torch.cat(\n",
    "            (torch.transpose(self.hadamard_product._indices(),0,1)\n",
    "             ,torch.tensor(self.extra_edges)),dim=0)\n",
    "        # Some edges from the 'extra_edges' and those given by the Hadamard product\n",
    "        # are duplicated. We need to combine these in one variable.   \n",
    "        self.edges_to_collapse_as_pairs = torch.unique(self.edges_to_collapse_as_pairs, dim=0)\n",
    "        self.all_nodes_to_identify  = torch.cat((rows,columns),dim=0).unique()\n",
    "        self._equivalenceclasses    = dict()\n",
    "        self.appendage_index        = len(self.seed_graph.nodes())\n",
    "        self.edge_index             = 0        \n",
    "        edge_pairs                  = torch.stack(self.seed_graph.edges(), dim = 1).int()\n",
    "        self.all_edges_as_pairs     = edge_pairs\n",
    "        self.edges_carry_fwd        = list()\n",
    "        self.edges_never_contracted = None\n",
    "        self.simplex_id             = 0\n",
    "        self._all_sets_with_indices = dict()\n",
    "        # Find all node classes to yield maximum class size.\n",
    "        self._globes = {element:partition for partition in relation(\n",
    "            self.edges_to_collapse_as_pairs) for element in partition}\n",
    "        index = 0\n",
    "        for key in self._globes.keys():\n",
    "            key_check = self._all_sets_with_indices.get(id(self._globes[key]),[])\n",
    "            if key_check == []:\n",
    "                self._all_sets_with_indices.update({id(self._globes[key]):(self._globes[key],index)})\n",
    "                index = index + 1\n",
    "        self.max_globe_index = index-1\n",
    "        index = None\n",
    "        self.maximum_class_size = max((len(set_value) for set_value in self._globes.values()), default=0)\n",
    "        self.loop_indicator     = torch.eq(self.srcs_and_dsts[0],self.srcs_and_dsts[1])\n",
    "        self.existing_loops     = (self.seed_graph.edges()[0][self.loop_indicator],\n",
    "                                   self.seed_graph.edges()[1][self.loop_indicator])\n",
    "        self.existing_loops     = torch.stack(self.existing_loops, dim = 1).int()\n",
    "        \n",
    "        # Finds globally highest dimension of simplices in the graph\n",
    "        self.in_degrees  = self.seed_graph.in_degrees()\n",
    "        self.out_degrees = self.seed_graph.out_degrees()\n",
    "        potential_max    = min(int(torch.max(self.in_degrees)),\n",
    "                               int(torch.max(self.out_degrees)),self.maximum_class_size)\n",
    "        if self.maximum_dimension > potential_max:\n",
    "            print(\"The given graph cannot have simplices of dimension\", self.maximum_dimension)\n",
    "            print(\"Changing \",self.maximum_dimension, \"to\", potential_max)\n",
    "            self.maximum_dimension = potential_max\n",
    "            \n",
    "    def _close_db(self):\n",
    "        self.connection.commit()\n",
    "        self.connection.close()\n",
    "        \n",
    "    def _connect_db(self):\n",
    "        self.connection = sqlite3.connect(self.database_name)\n",
    "        self.cursor = self.connection.cursor()\n",
    "        \n",
    "    def _view_db(self,table_name):\n",
    "        \"\"\"This function was created for easier visualization of the database\"\"\"\n",
    "        self._connect_db()\n",
    "        self.cursor.execute(f\"PRAGMA table_info({table_name})\")\n",
    "        if table_name == 'edge_details':\n",
    "            columns_info = self.cursor.fetchall()\n",
    "            reduced_columns = columns_info[1:10] \n",
    "            columns = [column[1] for column in reduced_columns]\n",
    "            self.cursor.execute(f\"SELECT {', '.join(columns)} FROM {table_name}\")\n",
    "            rows = self.cursor.fetchall()\n",
    "        else:\n",
    "            self.cursor.execute(f\"PRAGMA table_info({table_name})\")\n",
    "            columns = [column[1] for column in self.cursor.fetchall()]\n",
    "            self.cursor.execute(f\"SELECT * FROM {table_name}\")\n",
    "            rows = self.cursor.fetchall()\n",
    "        \n",
    "        table = tabulate(rows, headers=columns, tablefmt=\"pretty\")\n",
    "        print(table)\n",
    "        self._close_db()\n",
    "        \n",
    "        \n",
    "    def create_table(self):\n",
    "        \"\"\"This function creates the table, but does not intialize\n",
    "        data. Therefore, this function only needs to be run the first time. \"\"\"\n",
    "        self._connect_db()\n",
    "\n",
    "        self.cursor.execute('''\n",
    "            CREATE TABLE IF NOT EXISTS graph (\n",
    "                quotient_number INTEGER PRIMARY KEY,\n",
    "                number_of_nodes INTEGER,\n",
    "                number_of_edges INTEGER,\n",
    "                number_of_simplices INTEGER,\n",
    "                FOREIGN KEY (number_of_nodes) REFERENCES node_classes(node_class_id),\n",
    "                FOREIGN KEY (number_of_edges) REFERENCES edge_details(edge_id),\n",
    "                FOREIGN KEY (number_of_simplices) REFERENCES simplices(simplex_id)\n",
    "                                            )\n",
    "                            ''')\n",
    "        self.cursor.execute(f'''\n",
    "            CREATE TABLE IF NOT EXISTS node_classes (\n",
    "                node_class_id INTEGER PRIMARY KEY,\n",
    "                quotient_id INTEGER,\n",
    "                number_of_nodes INTEGER,\n",
    "                globe_id,\n",
    "                {', '.join(f'node_{i} INTEGER DEFAULT NULL' for i in range(0, self.maximum_class_size))}\n",
    "                                                    )\n",
    "                            ''')\n",
    "        self.cursor.execute('''\n",
    "            CREATE TABLE IF NOT EXISTS edge_details (\n",
    "                edge_id INTEGER PRIMARY KEY,\n",
    "                e_src INTEGER,\n",
    "                e_dst INTEGER,\n",
    "                edge_changed BOOLEAN,\n",
    "                to_contract BOOLEAN,\n",
    "                sampled BOOLEAN,\n",
    "                contracted BOOLEAN,\n",
    "                quotient_id INTEGER,\n",
    "                multiplicity INTEGER,\n",
    "                globe_id INTEGER,\n",
    "                number_of_edges INTEGER\n",
    "                                                    )\n",
    "                            ''')\n",
    "        self.cursor.execute(f'''\n",
    "            CREATE TABLE IF NOT EXISTS simplices (\n",
    "                simplex_id INTEGER PRIMARY KEY,\n",
    "                quotient_id INTEGER,\n",
    "                dimension INTEGER,\n",
    "                globe_id,\n",
    "                number_of_simplices INTEGER,\n",
    "                {', '.join(f'index_{i} INTEGER DEFAULT NULL' for i in range(0, self.maximum_class_size))}\n",
    "                                                    )\n",
    "                            ''')\n",
    "        self._close_db()\n",
    "        \n",
    "        \n",
    "    def initial_db_fill(self):\n",
    "        \"\"\"Fills in the database with the details from the (unquotiented) graph itself\n",
    "        i.e., before the quotienting process. \"\"\"\n",
    "        self._connect_db()\n",
    "        self.cursor.execute('''INSERT INTO graph \n",
    "                            (quotient_number, number_of_nodes, number_of_edges, \n",
    "                            number_of_simplices) VALUES\n",
    "                            (0, ?, ?, 0)\n",
    "                            ''', (len(self.seed_graph.nodes()), \n",
    "                                  len(self.seed_graph.edges()[0]))\n",
    "                           )\n",
    "        \n",
    "        for node in self.seed_graph.nodes():\n",
    "            picked_set = self._globes.get(int(node),{node})\n",
    "            picked_id  = self._all_sets_with_indices.get(id(picked_set),'X') \n",
    "            if picked_id == 'X':\n",
    "                globe_id = -1\n",
    "            else:\n",
    "                globe_id = picked_id[1]\n",
    "            self.cursor.execute('''INSERT INTO node_classes\n",
    "                                (node_class_id, node_0, quotient_id, \n",
    "                                number_of_nodes, globe_id) VALUES \n",
    "                                (?, ?, 0, ?, ?)\n",
    "                                ''', (int(node), int(node),  \n",
    "                                      len(self.seed_graph.nodes()),\n",
    "                                      globe_id))\n",
    "        for edge in self.all_edges_as_pairs:\n",
    "            if torch.any(torch.all(self.edges_to_collapse_as_pairs == edge, dim=1)):\n",
    "                picked_set = self._globes.get(int(edge[0]),{int(edge[0])})\n",
    "                picked_id  = self._all_sets_with_indices[id(picked_set)] \n",
    "                globe_id   = picked_id[1]\n",
    "\n",
    "                self.cursor.execute('''INSERT INTO edge_details\n",
    "                                    (edge_id, e_src, e_dst, edge_changed, \n",
    "                                    to_contract, quotient_id, multiplicity, sampled,\n",
    "                                    number_of_edges, contracted, globe_id) VALUES \n",
    "                                    (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)\n",
    "                                    ''', (self.edge_index, int(edge[0]), int(edge[1]), \n",
    "                                      False, True, 0, 1, False, len(self.seed_graph.edges()[0]),\n",
    "                                          False, globe_id)\n",
    "                                   )\n",
    "            elif torch.any(torch.all(self.existing_loops == edge, dim=1)):\n",
    "                self.cursor.execute('''INSERT INTO edge_details\n",
    "                                    (edge_id, e_src, e_dst, edge_changed, \n",
    "                                    to_contract, quotient_id, multiplicity, sampled,\n",
    "                                    number_of_edges, contracted, globe_id) VALUES \n",
    "                                    (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)\n",
    "                                    ''', (self.edge_index, int(edge[0]), int(edge[1]), \n",
    "                                      False, False, 0, 1, False, len(self.seed_graph.edges()[0]),\n",
    "                                          True, -1)\n",
    "                                   )\n",
    "\n",
    "            else:\n",
    "                self.cursor.execute('''INSERT INTO edge_details\n",
    "                                    (edge_id, e_src, e_dst, edge_changed, \n",
    "                                    to_contract, quotient_id, multiplicity, sampled,\n",
    "                                    number_of_edges, contracted, globe_id) VALUES \n",
    "                                    (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)\n",
    "                                    ''', (self.edge_index, int(edge[0]), int(edge[1]), \n",
    "                                      False, False, 0, 1, False, len(self.seed_graph.edges()[0]),\n",
    "                                         False, -1)\n",
    "                                   )\n",
    "            self.edge_index = self.edge_index + 1\n",
    "        \n",
    "        self._close_db()\n",
    "        \n",
    "    def sampling(self):\n",
    "        \"\"\"Samples edges to collapse, keeps track of changes in db for edges remaining to collapse.\"\"\"\n",
    "        self._connect_db()\n",
    "        # This query returns edges that have not yet been contracted.\n",
    "        self.cursor.execute(f'''SELECT e_src, e_dst\n",
    "                            FROM edge_details \n",
    "                            WHERE to_contract = True AND sampled = False\n",
    "                            AND edge_changed = False\n",
    "                            AND contracted = False AND quotient_id = {self.quotient_number}\n",
    "                            ORDER BY RANDOM()\n",
    "                            LIMIT {math.ceil(self.ratio * len(self.edges_to_collapse_as_pairs))}\n",
    "                            ''')\n",
    "        \n",
    "        self.selected_edges = self.cursor.fetchall()\n",
    "        # If two edges collapsing result in a third edge to collapse, we need to find it.\n",
    "        self._equivalenceclasses = {element:partition \n",
    "                                    for partition in relation(self.selected_edges) \n",
    "                                    for element in partition}\n",
    "        # When an edge collapses, we assign a new node label. The following list keeps track of this information.\n",
    "        changed_nodes = list(self._equivalenceclasses.keys())\n",
    "        \"\"\"To update entries of sampled edges in db, we first find all\n",
    "        the unique classes constructed above.\"\"\"\n",
    "        unique_sets = set()\n",
    "        for value in self._equivalenceclasses.values():\n",
    "            unique_sets.add(frozenset(value))\n",
    "        selected_edges = []\n",
    "        \n",
    "        for equivalence_class in unique_sets:\n",
    "            \"\"\"updates the column entry `sampled' for edges that have been sampled and those\n",
    "            that naturally won't be available for future quotienting\"\"\"\n",
    "            self.cursor.execute('''UPDATE edge_details\n",
    "                                SET sampled = True\n",
    "                                WHERE e_src IN ({}) AND e_dst IN ({}) AND quotient_id = ?\n",
    "                                '''.format(','.join(['?']*len(equivalence_class)),\n",
    "                                           ','.join(['?']*len(equivalence_class))),\n",
    "                                list(equivalence_class) + list(equivalence_class) +\n",
    "                                [self.quotient_number]\n",
    "                               )\n",
    "            \n",
    "            self.cursor.execute('''SELECT e_src, e_dst\n",
    "                                FROM edge_details\n",
    "                                WHERE contracted = False AND\n",
    "                                globe_id IS NOT -1 AND\n",
    "                                e_src IN ({}) AND e_dst IN ({})\n",
    "                                AND quotient_id = ?\n",
    "                                '''.format(','.join(['?']*len(equivalence_class)), \n",
    "                                            ','.join(['?']*len(equivalence_class))),\n",
    "                                    list(equivalence_class) + list(equivalence_class) +\n",
    "                               [self.quotient_number]\n",
    "                               )\n",
    "            selected_edges = selected_edges + self.cursor.fetchall()\n",
    "        self.selected_edges = selected_edges\n",
    "\n",
    "        \"\"\"Find out edges to carry forward\"\"\"\n",
    "        self.cursor.execute(f'''SELECT e_src, e_dst\n",
    "                            FROM edge_details \n",
    "                            WHERE to_contract = True AND sampled = False \n",
    "                            AND edge_changed = False AND contracted = False\n",
    "                            AND quotient_id ={self.quotient_number}\n",
    "                            ''')\n",
    "        \n",
    "        \"\"\"These are the egdes we do  not collapse after sampling and considering transitive closure\"\"\"\n",
    "        self.edges_carry_fwd = self.cursor.fetchall()\n",
    "        \n",
    "        self.cursor.execute('''UPDATE edge_details\n",
    "                            SET edge_changed = True\n",
    "                            WHERE e_src IN ({}) AND e_dst IN ({});\n",
    "                            '''.format(','.join(['?']*len(changed_nodes)),\n",
    "                                       ','.join(['?']*len(changed_nodes))), \n",
    "                            changed_nodes + changed_nodes)\n",
    "        self._close_db()\n",
    "        \n",
    "    def make_quotient(self):\n",
    "        \"\"\"This section can be combined with the db fill function above\n",
    "        The only reason this is kept distinct is to keep the class modular\"\"\"\n",
    "        self.new_edges_added = list()\n",
    "        self.new_nodes_added = set()\n",
    "        \n",
    "        self._sets = {id(self._equivalenceclasses[key]):self._equivalenceclasses[key]\n",
    "                      for key in self._equivalenceclasses.keys()}\n",
    "        # Create mapping of class names\n",
    "        self._classesnamesmapping = dict()\n",
    "        \n",
    "        for setid in self._sets.keys():\n",
    "            self._classesnamesmapping[setid] = self.appendage_index\n",
    "            self.appendage_index = self.appendage_index + 1\n",
    "                    \n",
    "        for key, value in self._classesnamesmapping.items():\n",
    "            element_of_set = next(iter(self._sets[key]))\n",
    "            globe_associated = self._globes[element_of_set]\n",
    "            self._globes.update({value:globe_associated})\n",
    "                    \n",
    "        # Each set (node class) is assigned a new node label.\n",
    "        for edge in self.selected_edges:\n",
    "            nodeclass    = self._equivalenceclasses[edge[0]]\n",
    "            newnodelabel = self._classesnamesmapping[id(nodeclass)]\n",
    "            self.new_edges_added = self.new_edges_added + [(newnodelabel,newnodelabel)]\n",
    "            self.new_nodes_added = self.new_nodes_added.union({newnodelabel})\n",
    "\n",
    "        # Since the node labels are changed, the edges that still need to be contracted will\n",
    "        # have their src and dst changed. To keep track of these un-contracted edges, we put them in the \n",
    "        # edges_carry_fwd variable\n",
    "        for edge in self.edges_carry_fwd:\n",
    "            srcnodeclass    = self._equivalenceclasses.get(edge[0], edge[0])\n",
    "            srcnewnodelabel = self._classesnamesmapping.get(id(srcnodeclass),edge[0])\n",
    "            dstnodeclass    = self._equivalenceclasses.get(edge[1], edge[1])\n",
    "            dstnewnodelabel = self._classesnamesmapping.get(id(dstnodeclass),edge[1])\n",
    "            self.new_edges_added = self.new_edges_added + [(srcnewnodelabel,dstnewnodelabel)]\n",
    "            self.new_nodes_added = self.new_nodes_added.union({srcnewnodelabel,dstnewnodelabel})\n",
    "            \n",
    "        print(\"Created a quotient\")\n",
    "        \n",
    "        # Remove variables to save space\n",
    "        self.edges_carry_fwd = None\n",
    "        self.selected_edges  = None\n",
    "        \n",
    "        self.quotient_number = self.quotient_number + 1\n",
    "        \n",
    "    def db_fill(self):\n",
    "        \"\"\"This function saves the details of a quotiented graph in the database.\n",
    "        Therefore, this function should be called immediately after the quotient is made.\n",
    "        All the edges of the graph are added, and are given different IDs, even if the edges\n",
    "        are already present in a previous quotient. However, we have different collections\n",
    "        here to ensure that the edges that have been changed are discarded from the edges\n",
    "        that need to be sampled. To this end, we add three new columns viz. to_contract, sampled\n",
    "        and contracted.\"\"\"\n",
    "        # To count the multiplicities of the edges\n",
    "        edge_counter        = collections.Counter(self.new_edges_added)\n",
    "        number_of_edges     = len(edge_counter)\n",
    "        number_of_nodes     = len(self.new_nodes_added)\n",
    "        \n",
    "        self._connect_db()\n",
    "        self.cursor.execute('''INSERT INTO graph \n",
    "                            (quotient_number, number_of_nodes, number_of_edges, \n",
    "                            number_of_simplices) VALUES\n",
    "                            (?, ?, ?, 0)\n",
    "                            ''', (self.quotient_number, \n",
    "                                  number_of_nodes,\n",
    "                                  number_of_edges)\n",
    "                           )\n",
    "        \n",
    "        edge_details_data = list()\n",
    "        for edge, count in edge_counter.items():\n",
    "            picked_set = self._globes[int(edge[0])]\n",
    "            picked_id = self._all_sets_with_indices[id(picked_set)]\n",
    "            globe_id = picked_id[1]\n",
    "            values = (self.edge_index, int(edge[0]), int(edge[1]),\n",
    "                      False, True, self.quotient_number, count,\n",
    "                      edge[0] == edge[1], number_of_edges,\n",
    "                      edge[0] == edge[1], globe_id)\n",
    "            edge_details_data.append(values)\n",
    "            self.edge_index += 1\n",
    "            \n",
    "        query = '''INSERT INTO edge_details \n",
    "                    (edge_id, e_src, e_dst, edge_changed, \n",
    "                    to_contract, quotient_id, multiplicity, sampled,\n",
    "                    number_of_edges, contracted, globe_id) VALUES \n",
    "                    (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)'''\n",
    "\n",
    "        self.cursor.executemany(query, edge_details_data)\n",
    "            \n",
    "        # Removing variable to save space\n",
    "        self.new_edges_added = None\n",
    "        edge_details_data = None\n",
    "        \n",
    "        node_classes_data = []\n",
    "\n",
    "        for setid, classlabel in self._classesnamesmapping.items():\n",
    "            node_data = self._sets[setid]\n",
    "            node = next(iter(node_data))\n",
    "            picked_set = self._globes.get(int(node),{node})\n",
    "            picked_id  = self._all_sets_with_indices.get(id(picked_set),'X')\n",
    "            if picked_id == 'X':\n",
    "                globe_id = -1\n",
    "            else:\n",
    "                globe_id = picked_id[1]\n",
    "            data_row = [globe_id, classlabel, self.quotient_number, number_of_nodes] + list(node_data) + [None] * (self.maximum_class_size - len(node_data))\n",
    "            node_classes_data.append(data_row)\n",
    "            \n",
    "        columns = ['globe_id', 'node_class_id', 'quotient_id', 'number_of_nodes']\n",
    "        columns = columns + [f'node_{i}' for i in range(0, self.maximum_class_size)]\n",
    "        placeholders = ', '.join(columns)\n",
    "        query = f'''INSERT INTO node_classes ({placeholders})\n",
    "                    VALUES ({', '.join(['?'] * len(columns))})\n",
    "                    '''\n",
    "        self.cursor.executemany(query, node_classes_data)\n",
    "        self._close_db()\n",
    "        \n",
    "                \n",
    "    def empty_table(self, table_name):\n",
    "        \"\"\"Made this function to empty database without deleting database. \n",
    "        Helpful for debugging\"\"\"\n",
    "        self._connect_db()\n",
    "        self.cursor.execute(f\"DELETE FROM {table_name}\")\n",
    "        self._close_db()\n",
    "        \n",
    "    def create_towers(self):\n",
    "        self.initial_db_fill()\n",
    "        print(\"DB initiliazed. Now creating quotients\")\n",
    "        for index in tqdm(range(self.bottom_level+1)):\n",
    "            self.sampling()\n",
    "            if len(self.selected_edges) == 0:\n",
    "                print(\"The graph cannot be quotiented further.\")\n",
    "                print(\"Changing bottom level to\",self.quotient_number)\n",
    "                self.bottom_level = self.quotient_number                \n",
    "                break\n",
    "            self.make_quotient()\n",
    "            self.db_fill()\n",
    "        print(\"Initial DB filled.\")\n",
    "            \n",
    "    def extract_from_db(self,quotient_number, globe_number):\n",
    "        \"\"\"This function extracts information about the graphs upstairs and downstairs, \n",
    "        given the quotient number and globe number. In addition, it also picks up simplices \n",
    "        from a graph below it in the hierarchy, and computes the lift of each simplex.\n",
    "        These lifts are fed to the simplicial search function Unpacker\"\"\"\n",
    "        \n",
    "        domain_graph = dgl.heterograph({('node', 'to', 'node'): ([], [])})\n",
    "                \n",
    "        self._connect_db()\n",
    "        # The constraint contracted = False ensures that we don't pick up loops.\n",
    "        self.cursor.execute(f'''SELECT e_src, e_dst\n",
    "                            FROM edge_details\n",
    "                            WHERE quotient_id = {quotient_number-1} \n",
    "                            AND contracted = False\n",
    "                            AND globe_id = {globe_number}\n",
    "                            ''')\n",
    "        domain_graph_edges = self.cursor.fetchall()\n",
    "        \n",
    "        self.cursor.execute(f'''SELECT * FROM simplices\n",
    "                            WHERE quotient_id = {quotient_number}\n",
    "                            AND globe_id = {globe_number}\n",
    "                            ''')\n",
    "        simplex_details = self.cursor.fetchall()\n",
    "        self.cursor.execute(f'''SELECT simplex_id FROM simplices\n",
    "                            ''')\n",
    "        all_simplex_ids = self.cursor.fetchall()\n",
    "        self.cursor.execute(f'''SELECT e_src, e_dst\n",
    "                            FROM edge_details\n",
    "                            WHERE quotient_id = {quotient_number}\n",
    "                            AND globe_id = {globe_number}\n",
    "                            ''')\n",
    "        range_graph_edges_db = self.cursor.fetchall()\n",
    "        self._close_db()\n",
    "        nodes_to_lift = set()\n",
    "        simplices_to_lift = dict()\n",
    "        if len(domain_graph_edges) == 0:\n",
    "            return ('empty_graph',)\n",
    "        \n",
    "        deg_edges = list()\n",
    "        range_graph_edges=list()\n",
    "        for row in range_graph_edges_db:\n",
    "            e_src        = row[0]\n",
    "            e_dst        = row[1]\n",
    "            nodes_to_lift = nodes_to_lift.union({e_src, e_dst})\n",
    "            if e_src == e_dst:\n",
    "                deg_edges = deg_edges + [[e_src, e_dst]]\n",
    "            else:\n",
    "                range_graph_edges.append([e_src,e_dst])\n",
    "        \n",
    "        nodes_to_lift = list(nodes_to_lift)\n",
    "        domain_graph_nodes = set()\n",
    "                \n",
    "        for src, dst in domain_graph_edges:\n",
    "            domain_graph.add_edges(src,dst)\n",
    "            domain_graph_nodes=domain_graph_nodes.union({src,dst})\n",
    "            \n",
    "        prev_nodes = list()\n",
    "        dimensions = list()\n",
    "\n",
    "        for row in simplex_details:\n",
    "            simplex_id, dimension = row[0], row[2]\n",
    "            index_values = row[5:]\n",
    "            index_values = list(value for value in index_values if value is not None)\n",
    "            simplices_to_lift[(dimension, simplex_id)] = index_values\n",
    "            if dimension == 0:\n",
    "                prev_nodes = prev_nodes + [row[4]]\n",
    "            dimensions = dimensions + [dimension]\n",
    "        simplex_index = max(all_simplex_ids, key=lambda x: x[0], default=(0,))[0] + 1\n",
    "        \n",
    "        \"\"\"Node IDs are kept the same as their labels. Since a node may be present in different quotient, \n",
    "        the nodes for each quotient are not present in the database. This information is extracted when\n",
    "        edges are constructed..\"\"\"\n",
    "        \n",
    "        \"\"\"Construct a dictionary of pre-images of the quotient ----> quotient+1 map on nodes. That is,\n",
    "        creates a mapping of node labels to their equivalence classes.\"\"\"\n",
    "\n",
    "        columns = ['node_class_id'] + ['globe_id'] +[f'node_{i}' for i in range(0, self.maximum_class_size)]\n",
    "        column_names = ', '.join(columns)\n",
    "        in_placeholders = ', '.join(['?' for _ in nodes_to_lift])\n",
    "        self._connect_db()\n",
    "        self.cursor.execute(f'''SELECT {column_names}\n",
    "                            FROM node_classes\n",
    "                            WHERE node_class_id IN ({in_placeholders})\n",
    "                            ''', tuple(nodes_to_lift))\n",
    "        node_mapping_db = self.cursor.fetchall()\n",
    "        \n",
    "        self._close_db()\n",
    "        \n",
    "        node_mapping  = dict()\n",
    "        \n",
    "        for row in node_mapping_db:\n",
    "            key      = row[0]\n",
    "            globe_id = row[1] \n",
    "            values   = row[2:]\n",
    "            values   = list(value for value in values if value is not None)\n",
    "            node_mapping.update({key: values})\n",
    "            \n",
    "        for node in list(domain_graph_nodes):\n",
    "            node_mapping[node] = [node]\n",
    "            \n",
    "        return (domain_graph, node_mapping, range_graph_edges, \n",
    "                deg_edges, simplices_to_lift, simplex_index,domain_graph_edges)\n",
    "        \n",
    "    def simplices_of_quotient(self,quotient_number):\n",
    "        \n",
    "        for globe_id in range(self.max_globe_index+1):\n",
    "            simplices = dict()\n",
    "            returns = self.extract_from_db(quotient_number, globe_id)\n",
    "            if returns[0] == 'empty_graph':\n",
    "                # We have an empty graph, so we can move on to the next iteration\n",
    "                continue\n",
    "            simplex_finder = Unpacker(domain_graph=returns[0],node_mapping=returns[1],range_graph_edges=returns[2],\n",
    "                                      deg_edges=returns[3],quotiented_simplices=returns[4],\n",
    "                                      simplex_max_index=returns[5],globe_number=globe_id,domain_graph_edges=returns[6])\n",
    "            if simplex_finder.local_max_dim == 1:\n",
    "                # There's no simplices to be found, so we can move on to the next iteration\n",
    "                continue\n",
    "            simplex_finder.inductive_connecting()\n",
    "            local_simplices = simplex_finder.lifted_simplices\n",
    "            simplices.update(local_simplices)\n",
    "            number_of_simplices = len(simplices.values())\n",
    "            if number_of_simplices == 0:\n",
    "                # No simplices have been found, so we can move on to the next iteration\n",
    "                continue\n",
    "            \n",
    "            # Otherwise, we add the found simplices to the database\n",
    "            \n",
    "            data_to_insert = []\n",
    "            for key, element in simplices.items():\n",
    "                dimension  = key[0]\n",
    "                simplex_id = key[1]\n",
    "                globe_id   = key[2]\n",
    "\n",
    "                data_row = [dimension, simplex_id, quotient_number-1, number_of_simplices,\n",
    "                            globe_id] + element + [None] * (self.maximum_class_size - len(element))\n",
    "                data_to_insert.append(data_row)\n",
    "            columns = ['dimension', 'simplex_id', 'quotient_id', 'number_of_simplices']\n",
    "            columns = columns + ['globe_id'] +[f'index_{i}' for i in range(self.maximum_class_size)]\n",
    "            placeholders = ', '.join(columns)\n",
    "            query = f'''INSERT INTO simplices ({placeholders})\n",
    "                    VALUES ({', '.join(['?'] * len(columns))})\n",
    "                    '''\n",
    "            self._connect_db()\n",
    "            self.cursor.executemany(query, data_to_insert)\n",
    "            self._close_db()\n",
    "        \n",
    "    def simplicial_search(self):        \n",
    "        print(\"Searching for simplices..\")\n",
    "        starting = self.bottom_level\n",
    "        for q_id in tqdm(range(starting, 0, -1)):\n",
    "            print(\"Search on-going in quotient number\",q_id)\n",
    "            self.simplices_of_quotient(q_id)\n",
    "            \n",
    "            \n",
    "\"\"\"found from https://stackoverflow.com/questions/42069187/\n",
    "create-a-list-of-unique-numbers-by-applying-transitive-closure\"\"\"\n",
    "def relation(array):\n",
    "\n",
    "    mapping = {}\n",
    "\n",
    "    def parent(u):\n",
    "        if mapping[u] == u:\n",
    "            return u\n",
    "        mapping[u] = parent(mapping[u])\n",
    "        return mapping[u]\n",
    "\n",
    "    for u, v in array:\n",
    "        u = int(u)\n",
    "        v = int(v)\n",
    "        if u not in mapping:\n",
    "            mapping[u] = u\n",
    "        if v not in mapping:\n",
    "            mapping[v] = v\n",
    "        mapping[parent(u)] = parent(v)\n",
    "\n",
    "    results = collections.defaultdict(set)\n",
    "    \n",
    "\n",
    "    for u in mapping.keys():\n",
    "        results[parent(u)].add(u)\n",
    "\n",
    "        \n",
    "    return [x for x in results.values()]\n",
    "\n",
    "            \n",
    "def find_common_tensors(tensor_A,tensor_B):\n",
    "    equal_pairs = torch.all(tensor_A[:, None, :] == tensor_B[None, :, :], dim=2)\n",
    "    common_pair_indices = torch.nonzero(equal_pairs, as_tuple=False)\n",
    "    return tensor_A[common_pair_indices[:, 0]]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "81dbc74c",
   "metadata": {},
   "outputs": [],
   "source": [
    "class SimplexCreator():\n",
    "    \"\"\"Create standard simplex\"\"\"\n",
    "    def __init__(self, dimension):\n",
    "        self.input_dimension = dimension\n",
    "        self.src=list()\n",
    "        self.dst=list()\n",
    "        for i in range(self.input_dimension+1):\n",
    "            for j in range(self.input_dimension+1):\n",
    "                if (i < j):\n",
    "                    self.src = self.src + [i]\n",
    "                    self.dst = self.dst + [j]"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "363ae0df",
   "metadata": {},
   "source": [
    "#### Code testing\n",
    "When running the graphs for the first time, you will need to run the create_table() module to set the database for this particular graph. Once that it done, ensure that you clear off the database using the empty_table() modules. The arguments have been fed already."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "98f91ce9",
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "C:\\Users\\abdul\\AppData\\Local\\Temp\\ipykernel_14412\\3455114506.py:16: UserWarning: Sparse CSR tensor support is in beta state. If you miss a functionality in the sparse tensor support, please submit a feature request to https://github.com/pytorch/pytorch/issues. (Triggered internally at ..\\aten\\src\\ATen\\SparseCsrTensorImpl.cpp:55.)\n",
      "  adj_squared = torch.sparse.mm(graph.adj_external(),graph.adj_external())\n"
     ]
    }
   ],
   "source": [
    "K_5 = dgl.heterograph({('paper', 'cites', 'paper'): (SimplexCreator(dimension=5).src, SimplexCreator(dimension=5).dst)})\n",
    "filepath = 'K_5'\n",
    "db = 'K_5.db'\n",
    "K_5_preprocessing = graph_towers(filepath,graph=K_5, database_name=db, ratio=0.01,bottom_level = 10, max_dimension = 23)\n",
    "K_5_preprocessing.create_table()\n",
    "#K_5_preprocessing.empty_table('edge_details')\n",
    "#K_5_preprocessing.empty_table('graph')\n",
    "#K_5_preprocessing.empty_table('node_classes')\n",
    "#K_5_preprocessing.empty_table('simplices')\n",
    "K_5_preprocessing.create_towers()\n",
    "K_5_preprocessing.simplicial_search()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "a9166e5f",
   "metadata": {},
   "outputs": [],
   "source": [
    "#cProfile.run('K_5_preprocessing.simplicial_search()')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "9e0bf0a7",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "| quotient_number | number_of_nodes | number_of_edges | number_of_simplices |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "|        0        |        6        |       15        |          0          |\n",
      "|        1        |        5        |       13        |          0          |\n",
      "|        2        |        4        |       10        |          0          |\n",
      "|        3        |        3        |        5        |          0          |\n",
      "|        4        |        2        |        3        |          0          |\n",
      "|        5        |        1        |        1        |          0          |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "| node_class_id | quotient_id | number_of_nodes | globe_id | node_0 | node_1 | node_2 | node_3 | node_4 | node_5 |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "|       0       |      0      |        6        |    0     |   0    |        |        |        |        |        |\n",
      "|       1       |      0      |        6        |    0     |   1    |        |        |        |        |        |\n",
      "|       2       |      0      |        6        |    0     |   2    |        |        |        |        |        |\n",
      "|       3       |      0      |        6        |    0     |   3    |        |        |        |        |        |\n",
      "|       4       |      0      |        6        |    0     |   4    |        |        |        |        |        |\n",
      "|       5       |      0      |        6        |    0     |   5    |        |        |        |        |        |\n",
      "|       6       |      1      |        5        |    0     |   1    |   4    |        |        |        |        |\n",
      "|       7       |      2      |        4        |    0     |   0    |   3    |        |        |        |        |\n",
      "|       8       |      3      |        3        |    0     |   6    |   7    |        |        |        |        |\n",
      "|       9       |      4      |        2        |    0     |   8    |   5    |        |        |        |        |\n",
      "|      10       |      5      |        1        |    0     |   9    |   2    |        |        |        |        |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "| e_src | e_dst | edge_changed | to_contract | sampled | contracted | quotient_id | multiplicity | globe_id |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   3   |      1       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   4   |      1       |      1      |    1    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   6   |   6   |      1       |      1      |    1    |     1      |      1      |      1       |    0     |\n",
      "|   0   |   6   |      0       |      1      |    0    |     0      |      1      |      2       |    0     |\n",
      "|   0   |   2   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   3   |      1       |      1      |    1    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   6   |   2   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   6   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   6   |   5   |      0       |      1      |    0    |     0      |      1      |      2       |    0     |\n",
      "|   2   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   2   |   6   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   3   |   6   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   7   |   7   |      1       |      1      |    1    |     1      |      2      |      1       |    0     |\n",
      "|   7   |   6   |      1       |      1      |    1    |     0      |      2      |      2       |    0     |\n",
      "|   7   |   2   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   7   |   5   |      0       |      1      |    0    |     0      |      2      |      2       |    0     |\n",
      "|   6   |   2   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   6   |   7   |      1       |      1      |    1    |     0      |      2      |      1       |    0     |\n",
      "|   6   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   2   |   7   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   2   |   6   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   8   |   8   |      1       |      1      |    1    |     1      |      3      |      2       |    0     |\n",
      "|   8   |   2   |      0       |      1      |    0    |     0      |      3      |      2       |    0     |\n",
      "|   8   |   5   |      1       |      1      |    1    |     0      |      3      |      2       |    0     |\n",
      "|   2   |   8   |      0       |      1      |    0    |     0      |      3      |      2       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   9   |   9   |      1       |      1      |    1    |     1      |      4      |      1       |    0     |\n",
      "|   9   |   2   |      1       |      1      |    1    |     0      |      4      |      1       |    0     |\n",
      "|   2   |   9   |      1       |      1      |    1    |     0      |      4      |      2       |    0     |\n",
      "|  10   |  10   |      0       |      1      |    1    |     1      |      5      |      2       |    0     |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "| simplex_id | quotient_id | dimension | globe_id | number_of_simplices | index_0 | index_1 | index_2 | index_3 | index_4 | index_5 |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "|     1      |      3      |     2     |    0     |          2          |    2    |    8    |    5    |         |         |         |\n",
      "|     2      |      3      |     2     |    0     |          2          |    8    |    2    |    5    |         |         |         |\n",
      "|     3      |      2      |     2     |    0     |         18          |    2    |    6    |    5    |         |         |         |\n",
      "|     4      |      2      |     2     |    0     |         18          |    2    |    7    |    5    |         |         |         |\n",
      "|     5      |      2      |     2     |    0     |         18          |    6    |    2    |    5    |         |         |         |\n",
      "|     6      |      2      |     2     |    0     |         18          |    7    |    2    |    5    |         |         |         |\n",
      "|     7      |      2      |     3     |    0     |         18          |    2    |    6    |    7    |    5    |         |         |\n",
      "|     8      |      2      |     3     |    0     |         18          |    2    |    7    |    6    |    5    |         |         |\n",
      "|     9      |      2      |     3     |    0     |         18          |    6    |    2    |    7    |    5    |         |         |\n",
      "|     10     |      2      |     3     |    0     |         18          |    7    |    2    |    6    |    5    |         |         |\n",
      "|     11     |      2      |     3     |    0     |         18          |    6    |    7    |    2    |    5    |         |         |\n",
      "|     12     |      2      |     3     |    0     |         18          |    7    |    6    |    2    |    5    |         |         |\n",
      "|     13     |      2      |     2     |    0     |         18          |    6    |    7    |    2    |         |         |         |\n",
      "|     14     |      2      |     2     |    0     |         18          |    7    |    6    |    2    |         |         |         |\n",
      "|     15     |      2      |     2     |    0     |         18          |    6    |    7    |    5    |         |         |         |\n",
      "|     16     |      2      |     2     |    0     |         18          |    7    |    6    |    5    |         |         |         |\n",
      "|     17     |      2      |     2     |    0     |         18          |    2    |    6    |    7    |         |         |         |\n",
      "|     18     |      2      |     2     |    0     |         18          |    2    |    7    |    6    |         |         |         |\n",
      "|     19     |      2      |     2     |    0     |         18          |    6    |    2    |    7    |         |         |         |\n",
      "|     20     |      2      |     2     |    0     |         18          |    7    |    2    |    6    |         |         |         |\n",
      "|     21     |      1      |     2     |    0     |         30          |    2    |    6    |    5    |         |         |         |\n",
      "|     22     |      1      |     2     |    0     |         30          |    2    |    3    |    5    |         |         |         |\n",
      "|     23     |      1      |     2     |    0     |         30          |    6    |    2    |    5    |         |         |         |\n",
      "|     24     |      1      |     2     |    0     |         30          |    0    |    2    |    5    |         |         |         |\n",
      "|     25     |      1      |     3     |    0     |         30          |    2    |    6    |    3    |    5    |         |         |\n",
      "|     26     |      1      |     3     |    0     |         30          |    2    |    3    |    6    |    5    |         |         |\n",
      "|     27     |      1      |     3     |    0     |         30          |    6    |    2    |    3    |    5    |         |         |\n",
      "|     28     |      1      |     3     |    0     |         30          |    0    |    2    |    6    |    5    |         |         |\n",
      "|     29     |      1      |     3     |    0     |         30          |    0    |    6    |    2    |    5    |         |         |\n",
      "|     30     |      1      |     2     |    0     |         30          |    0    |    6    |    2    |         |         |         |\n",
      "|     31     |      1      |     2     |    0     |         30          |    6    |    3    |    5    |         |         |         |\n",
      "|     32     |      1      |     2     |    0     |         30          |    0    |    6    |    5    |         |         |         |\n",
      "|     33     |      1      |     2     |    0     |         30          |    3    |    6    |    5    |         |         |         |\n",
      "|     34     |      1      |     2     |    0     |         30          |    2    |    6    |    3    |         |         |         |\n",
      "|     35     |      1      |     2     |    0     |         30          |    2    |    3    |    6    |         |         |         |\n",
      "|     36     |      1      |     2     |    0     |         30          |    6    |    2    |    3    |         |         |         |\n",
      "|     37     |      1      |     2     |    0     |         30          |    0    |    2    |    6    |         |         |         |\n",
      "|     38     |      1      |     3     |    0     |         30          |    0    |    2    |    3    |    5    |         |         |\n",
      "|     39     |      1      |     4     |    0     |         30          |    0    |    2    |    6    |    3    |    5    |         |\n",
      "|     40     |      1      |     4     |    0     |         30          |    0    |    2    |    3    |    6    |    5    |         |\n",
      "|     41     |      1      |     4     |    0     |         30          |    0    |    6    |    2    |    3    |    5    |         |\n",
      "|     42     |      1      |     3     |    0     |         30          |    0    |    6    |    3    |    5    |         |         |\n",
      "|     43     |      1      |     3     |    0     |         30          |    0    |    3    |    6    |    5    |         |         |\n",
      "|     44     |      1      |     3     |    0     |         30          |    0    |    2    |    6    |    3    |         |         |\n",
      "|     45     |      1      |     3     |    0     |         30          |    0    |    2    |    3    |    6    |         |         |\n",
      "|     46     |      1      |     3     |    0     |         30          |    0    |    6    |    2    |    3    |         |         |\n",
      "|     47     |      1      |     2     |    0     |         30          |    0    |    3    |    6    |         |         |         |\n",
      "|     48     |      1      |     2     |    0     |         30          |    0    |    3    |    5    |         |         |         |\n",
      "|     49     |      1      |     2     |    0     |         30          |    0    |    6    |    3    |         |         |         |\n",
      "|     50     |      1      |     2     |    0     |         30          |    0    |    2    |    3    |         |         |         |\n",
      "|     51     |      0      |     2     |    0     |         42          |    2    |    4    |    5    |         |         |         |\n",
      "|     52     |      0      |     2     |    0     |         42          |    2    |    3    |    5    |         |         |         |\n",
      "|     53     |      0      |     2     |    0     |         42          |    1    |    2    |    5    |         |         |         |\n",
      "|     54     |      0      |     2     |    0     |         42          |    0    |    2    |    5    |         |         |         |\n",
      "|     55     |      0      |     3     |    0     |         42          |    2    |    3    |    4    |    5    |         |         |\n",
      "|     56     |      0      |     3     |    0     |         42          |    1    |    2    |    3    |    5    |         |         |\n",
      "|     57     |      0      |     3     |    0     |         42          |    0    |    2    |    4    |    5    |         |         |\n",
      "|     58     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    5    |         |         |\n",
      "|     59     |      0      |     2     |    0     |         42          |    0    |    1    |    2    |         |         |         |\n",
      "|     60     |      0      |     2     |    0     |         42          |    1    |    3    |    5    |         |         |         |\n",
      "|     61     |      0      |     2     |    0     |         42          |    0    |    1    |    5    |         |         |         |\n",
      "|     62     |      0      |     2     |    0     |         42          |    0    |    4    |    5    |         |         |         |\n",
      "|     63     |      0      |     2     |    0     |         42          |    3    |    4    |    5    |         |         |         |\n",
      "|     64     |      0      |     2     |    0     |         42          |    2    |    3    |    4    |         |         |         |\n",
      "|     65     |      0      |     2     |    0     |         42          |    1    |    2    |    3    |         |         |         |\n",
      "|     66     |      0      |     2     |    0     |         42          |    0    |    2    |    4    |         |         |         |\n",
      "|     67     |      0      |     3     |    0     |         42          |    0    |    2    |    3    |    5    |         |         |\n",
      "|     68     |      0      |     4     |    0     |         42          |    0    |    2    |    3    |    4    |    5    |         |\n",
      "|     69     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    5    |         |\n",
      "|     70     |      0      |     3     |    0     |         42          |    0    |    1    |    3    |    5    |         |         |\n",
      "|     71     |      0      |     3     |    0     |         42          |    0    |    3    |    4    |    5    |         |         |\n",
      "|     72     |      0      |     3     |    0     |         42          |    0    |    2    |    3    |    4    |         |         |\n",
      "|     73     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    3    |         |         |\n",
      "|     74     |      0      |     2     |    0     |         42          |    0    |    3    |    4    |         |         |         |\n",
      "|     75     |      0      |     2     |    0     |         42          |    0    |    3    |    5    |         |         |         |\n",
      "|     76     |      0      |     2     |    0     |         42          |    0    |    1    |    3    |         |         |         |\n",
      "|     77     |      0      |     2     |    0     |         42          |    0    |    2    |    3    |         |         |         |\n",
      "|     78     |      0      |     3     |    0     |         42          |    1    |    2    |    4    |    5    |         |         |\n",
      "|     79     |      0      |     4     |    0     |         42          |    1    |    2    |    3    |    4    |    5    |         |\n",
      "|     80     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    4    |    5    |         |\n",
      "|     81     |      0      |     3     |    0     |         42          |    1    |    3    |    4    |    5    |         |         |\n",
      "|     82     |      0      |     3     |    0     |         42          |    0    |    1    |    4    |    5    |         |         |\n",
      "|     83     |      0      |     3     |    0     |         42          |    1    |    2    |    3    |    4    |         |         |\n",
      "|     84     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    4    |         |         |\n",
      "|     85     |      0      |     5     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |    5    |\n",
      "|     86     |      0      |     4     |    0     |         42          |    0    |    1    |    3    |    4    |    5    |         |\n",
      "|     87     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |         |\n",
      "|     88     |      0      |     3     |    0     |         42          |    0    |    1    |    3    |    4    |         |         |\n",
      "|     89     |      0      |     2     |    0     |         42          |    0    |    1    |    4    |         |         |         |\n",
      "|     90     |      0      |     2     |    0     |         42          |    1    |    4    |    5    |         |         |         |\n",
      "|     91     |      0      |     2     |    0     |         42          |    1    |    2    |    4    |         |         |         |\n",
      "|     92     |      0      |     2     |    0     |         42          |    1    |    3    |    4    |         |         |         |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n"
     ]
    }
   ],
   "source": [
    "# View the resulting databases after the search\n",
    "K_5_preprocessing._view_db('graph')\n",
    "K_5_preprocessing._view_db('node_classes')\n",
    "K_5_preprocessing._view_db('edge_details')\n",
    "K_5_preprocessing._view_db('simplices')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "69559856",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "         4623 function calls (4617 primitive calls) in 0.198 seconds\n",
      "\n",
      "   Ordered by: standard name\n",
      "\n",
      "   ncalls  tottime  percall  cumtime  percall filename:lineno(function)\n",
      "       14    0.002    0.000    0.014    0.001 279130135.py:101(_connect_db)\n",
      "        1    0.004    0.004    0.034    0.034 279130135.py:179(initial_db_fill)\n",
      "        7    0.001    0.000    0.090    0.013 279130135.py:245(sampling)\n",
      "        7    0.000    0.000    0.000    0.000 279130135.py:260(<dictcomp>)\n",
      "        6    0.001    0.000    0.001    0.000 279130135.py:317(make_quotient)\n",
      "        6    0.000    0.000    0.000    0.000 279130135.py:323(<dictcomp>)\n",
      "        6    0.001    0.000    0.073    0.012 279130135.py:363(db_fill)\n",
      "        6    0.000    0.000    0.000    0.000 279130135.py:425(<listcomp>)\n",
      "        1    0.001    0.001    0.198    0.198 279130135.py:441(create_towers)\n",
      "        7    0.000    0.000    0.000    0.000 279130135.py:620(relation)\n",
      "    30/24    0.000    0.000    0.000    0.000 279130135.py:624(parent)\n",
      "        7    0.000    0.000    0.000    0.000 279130135.py:646(<listcomp>)\n",
      "       14    0.000    0.000    0.105    0.008 279130135.py:97(_close_db)\n",
      "      100    0.000    0.000    0.001    0.000 <frozen abc>:117(__instancecheck__)\n",
      "       66    0.000    0.000    0.000    0.000 <frozen abc>:121(__subclasscheck__)\n",
      "        1    0.000    0.000    0.198    0.198 <string>:1(<module>)\n",
      "       42    0.000    0.000    0.003    0.000 __init__.py:145(from_dgl_nd)\n",
      "       84    0.000    0.000    0.000    0.000 __init__.py:511(cast)\n",
      "        6    0.000    0.000    0.000    0.000 __init__.py:587(__init__)\n",
      "        6    0.000    0.000    0.000    0.000 __init__.py:660(update)\n",
      "        8    0.000    0.000    0.000    0.000 _tensor.py:1002(__hash__)\n",
      "       30    0.000    0.000    0.001    0.000 _tensor.py:964(__len__)\n",
      "        2    0.000    0.000    0.000    0.000 _tensor.py:980(__iter__)\n",
      "       10    0.000    0.000    0.000    0.000 base.py:23(<lambda>)\n",
      "      240    0.000    0.000    0.000    0.000 base.py:58(check_call)\n",
      "       14    0.000    0.000    0.000    0.000 base.py:73(c_str)\n",
      "       42    0.000    0.000    0.001    0.000 dlpack.py:49(from_dlpack)\n",
      "      114    0.002    0.000    0.003    0.000 function.py:103(_make_dgl_args)\n",
      "       14    0.000    0.000    0.000    0.000 function.py:184(__init__)\n",
      "       14    0.000    0.000    0.000    0.000 function.py:198(__del__)\n",
      "      114    0.002    0.000    0.007    0.000 function.py:202(__call__)\n",
      "       14    0.000    0.000    0.000    0.000 function.py:258(_handle_return_func)\n",
      "       42    0.000    0.000    0.000    0.000 function.py:270(<lambda>)\n",
      "       14    0.000    0.000    0.001    0.000 functools.py:35(update_wrapper)\n",
      "       14    0.000    0.000    0.000    0.000 functools.py:479(lru_cache)\n",
      "       14    0.000    0.000    0.001    0.000 functools.py:518(decorating_function)\n",
      "       14    0.000    0.000    0.000    0.000 graph_index.py:174(num_edges)\n",
      "       14    0.000    0.000    0.000    0.000 graph_index.py:39(__new__)\n",
      "       10    0.000    0.000    0.000    0.000 heterograph.py:1297(get_ntype_id)\n",
      "       14    0.000    0.000    0.002    0.000 heterograph.py:1389(get_etype_id)\n",
      "       10    0.000    0.000    0.000    0.000 heterograph.py:1775(nodes)\n",
      "       14    0.000    0.000    0.000    0.000 heterograph.py:2154(edges)\n",
      "       10    0.000    0.000    0.001    0.000 heterograph.py:2831(idtype)\n",
      "       14    0.000    0.000    0.010    0.001 heterograph.py:3519(all_edges)\n",
      "       10    0.000    0.000    0.001    0.000 heterograph.py:5629(device)\n",
      "       20    0.000    0.000    0.000    0.000 heterograph.py:959(is_unibipartite)\n",
      "       10    0.000    0.000    0.001    0.000 heterograph_index.py:178(dtype)\n",
      "       10    0.000    0.000    0.000    0.000 heterograph_index.py:189(ctx)\n",
      "       10    0.000    0.000    0.001    0.000 heterograph_index.py:382(num_nodes)\n",
      "       14    0.000    0.000    0.007    0.001 heterograph_index.py:666(edges)\n",
      "       14    0.000    0.000    0.002    0.000 heterograph_index.py:74(metagraph)\n",
      "       14    0.000    0.000    0.002    0.000 heterograph_index.py:93(number_of_etypes)\n",
      "       42    0.000    0.000    0.000    0.000 ndarray.py:113(_make_array)\n",
      "      126    0.001    0.000    0.001    0.000 ndarray.py:206(shape)\n",
      "      252    0.000    0.000    0.000    0.000 ndarray.py:209(<genexpr>)\n",
      "       42    0.000    0.000    0.000    0.000 ndarray.py:48(_dlpack_deleter)\n",
      "       42    0.000    0.000    0.000    0.000 ndarray.py:70(__init__)\n",
      "       42    0.000    0.000    0.000    0.000 ndarray.py:81(__del__)\n",
      "       42    0.000    0.000    0.001    0.000 ndarray.py:89(to_dlpack)\n",
      "       14    0.000    0.000    0.000    0.000 object.py:28(_return_object)\n",
      "       14    0.000    0.000    0.000    0.000 object.py:55(__del__)\n",
      "       10    0.000    0.000    0.000    0.000 runtime_ctypes.py:151(__new__)\n",
      "       10    0.000    0.000    0.000    0.000 runtime_ctypes.py:163(__init__)\n",
      "       10    0.000    0.000    0.000    0.000 tensor.py:115(to_backend_ctx)\n",
      "       10    0.000    0.000    0.001    0.000 tensor.py:406(arange)\n",
      "       42    0.001    0.000    0.003    0.000 tensor.py:465(zerocopy_from_dgl_ndarray)\n",
      "       24    0.000    0.000    0.000    0.000 types.py:70(<lambda>)\n",
      "       10    0.000    0.000    0.000    0.000 types.py:74(<lambda>)\n",
      "       10    0.000    0.000    0.000    0.000 types.py:76(<lambda>)\n",
      "       14    0.000    0.000    0.000    0.000 view.py:147(__init__)\n",
      "       14    0.000    0.000    0.010    0.001 view.py:177(__call__)\n",
      "       10    0.000    0.000    0.000    0.000 view.py:20(__init__)\n",
      "       10    0.000    0.000    0.003    0.000 view.py:46(__call__)\n",
      "       14    0.000    0.000    0.000    0.000 {built-in method __new__ of type object at 0x00007FFB9675A4B0}\n",
      "      100    0.000    0.000    0.001    0.000 {built-in method _abc._abc_instancecheck}\n",
      "       66    0.000    0.000    0.000    0.000 {built-in method _abc._abc_subclasscheck}\n",
      "        6    0.000    0.000    0.000    0.000 {built-in method _collections._count_elements}\n",
      "      284    0.000    0.000    0.000    0.000 {built-in method _ctypes.byref}\n",
      "       14    0.010    0.001    0.011    0.001 {built-in method _sqlite3.connect}\n",
      "        1    0.000    0.000    0.198    0.198 {built-in method builtins.exec}\n",
      "      108    0.000    0.000    0.000    0.000 {built-in method builtins.getattr}\n",
      "       56    0.000    0.000    0.000    0.000 {built-in method builtins.hasattr}\n",
      "      148    0.000    0.000    0.000    0.000 {built-in method builtins.id}\n",
      "      590    0.000    0.000    0.001    0.000 {built-in method builtins.isinstance}\n",
      "       14    0.000    0.000    0.000    0.000 {built-in method builtins.iter}\n",
      "      265    0.000    0.000    0.001    0.000 {built-in method builtins.len}\n",
      "       42    0.000    0.000    0.000    0.000 {built-in method builtins.min}\n",
      "       12    0.000    0.000    0.000    0.000 {built-in method builtins.next}\n",
      "       28    0.000    0.000    0.000    0.000 {built-in method builtins.setattr}\n",
      "        7    0.000    0.000    0.000    0.000 {built-in method math.ceil}\n",
      "       42    0.000    0.000    0.000    0.000 {built-in method torch._C._from_dlpack}\n",
      "       32    0.000    0.000    0.000    0.000 {built-in method torch._C._get_tracing_state}\n",
      "       30    0.000    0.000    0.000    0.000 {built-in method torch._C._has_torch_function_unary}\n",
      "       14    0.001    0.000    0.001    0.000 {built-in method torch.all}\n",
      "       14    0.000    0.000    0.000    0.000 {built-in method torch.any}\n",
      "       10    0.001    0.000    0.001    0.000 {built-in method torch.arange}\n",
      "       24    0.000    0.000    0.000    0.000 {method 'add' of 'set' objects}\n",
      "       39    0.000    0.000    0.000    0.000 {method 'append' of 'list' objects}\n",
      "       14    0.004    0.000    0.004    0.000 {method 'close' of 'sqlite3.Connection' objects}\n",
      "       14    0.102    0.007    0.102    0.007 {method 'commit' of 'sqlite3.Connection' objects}\n",
      "       14    0.000    0.000    0.000    0.000 {method 'cursor' of 'sqlite3.Connection' objects}\n",
      "       10    0.000    0.000    0.000    0.000 {method 'decode' of 'bytes' objects}\n",
      "       32    0.000    0.000    0.000    0.000 {method 'dim' of 'torch._C._TensorBase' objects}\n",
      "        1    0.000    0.000    0.000    0.000 {method 'disable' of '_lsprof.Profiler' objects}\n",
      "       14    0.000    0.000    0.000    0.000 {method 'encode' of 'str' objects}\n",
      "       61    0.050    0.001    0.050    0.001 {method 'execute' of 'sqlite3.Cursor' objects}\n",
      "       12    0.005    0.000    0.005    0.000 {method 'executemany' of 'sqlite3.Cursor' objects}\n",
      "       20    0.000    0.000    0.000    0.000 {method 'fetchall' of 'sqlite3.Cursor' objects}\n",
      "       19    0.000    0.000    0.000    0.000 {method 'format' of 'str' objects}\n",
      "      174    0.000    0.000    0.000    0.000 {method 'get' of 'dict' objects}\n",
      "       18    0.000    0.000    0.000    0.000 {method 'items' of 'dict' objects}\n",
      "       50    0.000    0.000    0.000    0.000 {method 'join' of 'str' objects}\n",
      "       26    0.000    0.000    0.000    0.000 {method 'keys' of 'dict' objects}\n",
      "        2    0.000    0.000    0.000    0.000 {method 'unbind' of 'torch._C._TensorBase' objects}\n",
      "       39    0.000    0.000    0.000    0.000 {method 'union' of 'set' objects}\n",
      "       20    0.000    0.000    0.000    0.000 {method 'update' of 'dict' objects}\n",
      "       14    0.000    0.000    0.000    0.000 {method 'values' of 'dict' objects}\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "src = [0,0,0,1,1,2] + [1] + [4,4,4,5,5,6] \n",
    "dst = [1,2,3,2,3,3] + [4] + [5,6,7,6,7,7] \n",
    "twosimplices = dgl.heterograph({('paper', 'cites', 'paper'): (src, dst)})\n",
    "filepath = 'twosimplices'\n",
    "db2 = 'twosimplices.db'\n",
    "twosimplices_preprocessing = graph_towers(filepath,database_name = db2, graph=twosimplices,ratio=0.01,bottom_level = 20, max_dimension = 20)\n",
    "twosimplices_preprocessing.create_table()\n",
    "#twosimplices_preprocessing.empty_table('edge_details')\n",
    "#twosimplices_preprocessing.empty_table('graph')\n",
    "#twosimplices_preprocessing.empty_table('node_classes')\n",
    "#twosimplices_preprocessing.empty_table('simplices')\n",
    "twosimplices_preprocessing.create_towers()\n",
    "twosimplices_preprocessing.simplicial_search()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "df739a1c",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "| quotient_number | number_of_nodes | number_of_edges | number_of_simplices |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "|        0        |        8        |       13        |          0          |\n",
      "|        1        |        7        |       10        |          0          |\n",
      "|        2        |        6        |        9        |          0          |\n",
      "|        3        |        5        |        6        |          0          |\n",
      "|        4        |        4        |        4        |          0          |\n",
      "|        5        |        2        |        3        |          0          |\n",
      "|        6        |        1        |        1        |          0          |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+\n",
      "| node_class_id | quotient_id | number_of_nodes | globe_id | node_0 | node_1 | node_2 | node_3 |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+\n",
      "|       0       |      0      |        8        |    0     |   0    |        |        |        |\n",
      "|       1       |      0      |        8        |    0     |   1    |        |        |        |\n",
      "|       2       |      0      |        8        |    0     |   2    |        |        |        |\n",
      "|       3       |      0      |        8        |    0     |   3    |        |        |        |\n",
      "|       4       |      0      |        8        |    1     |   4    |        |        |        |\n",
      "|       5       |      0      |        8        |    1     |   5    |        |        |        |\n",
      "|       6       |      0      |        8        |    1     |   6    |        |        |        |\n",
      "|       7       |      0      |        8        |    1     |   7    |        |        |        |\n",
      "|       8       |      1      |        7        |    0     |   2    |   3    |        |        |\n",
      "|       9       |      2      |        6        |    1     |   4    |   7    |        |        |\n",
      "|      10       |      3      |        5        |    1     |   9    |   5    |        |        |\n",
      "|      11       |      4      |        4        |    1     |   10   |   6    |        |        |\n",
      "|      12       |      5      |        2        |    0     |   0    |   8    |        |        |\n",
      "|      13       |      6      |        1        |    0     |   1    |   12   |        |        |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "| e_src | e_dst | edge_changed | to_contract | sampled | contracted | quotient_id | multiplicity | globe_id |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   3   |      1       |      1      |    1    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   4   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   4   |   6   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   4   |   7   |      1       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   5   |   6   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   5   |   7   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   8   |   8   |      1       |      1      |    1    |     1      |      1      |      1       |    0     |\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   8   |      1       |      1      |    0    |     0      |      1      |      2       |    0     |\n",
      "|   1   |   8   |      0       |      1      |    0    |     0      |      1      |      2       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   4   |   6   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   4   |   7   |      1       |      1      |    1    |     0      |      1      |      1       |    1     |\n",
      "|   5   |   6   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   5   |   7   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   9   |   9   |      1       |      1      |    1    |     1      |      2      |      1       |    1     |\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   0   |   8   |      1       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   8   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   9   |   5   |      1       |      1      |    1    |     0      |      2      |      1       |    1     |\n",
      "|   9   |   6   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   5   |   6   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   5   |   9   |      1       |      1      |    1    |     0      |      2      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|  10   |  10   |      1       |      1      |    1    |     1      |      3      |      2       |    1     |\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   0   |   8   |      1       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   8   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|  10   |   6   |      1       |      1      |    1    |     0      |      3      |      2       |    1     |\n",
      "|   6   |  10   |      1       |      1      |    1    |     0      |      3      |      1       |    1     |\n",
      "|  11   |  11   |      0       |      1      |    1    |     1      |      4      |      2       |    1     |\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   0   |   8   |      1       |      1      |    1    |     0      |      4      |      1       |    0     |\n",
      "|   1   |   8   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|  12   |  12   |      1       |      1      |    1    |     1      |      5      |      1       |    0     |\n",
      "|  12   |   1   |      1       |      1      |    1    |     0      |      5      |      1       |    0     |\n",
      "|   1   |  12   |      1       |      1      |    1    |     0      |      5      |      1       |    0     |\n",
      "|  13   |  13   |      0       |      1      |    1    |     1      |      6      |      2       |    0     |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+\n",
      "| simplex_id | quotient_id | dimension | globe_id | number_of_simplices | index_0 | index_1 | index_2 | index_3 |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+\n",
      "|     1      |      4      |     2     |    0     |          1          |    0    |    1    |    8    |         |\n",
      "|     2      |      3      |     2     |    0     |          1          |    0    |    1    |    8    |         |\n",
      "|     3      |      2      |     2     |    0     |          1          |    0    |    1    |    8    |         |\n",
      "|     4      |      2      |     2     |    1     |          3          |    9    |    5    |    6    |         |\n",
      "|     5      |      2      |     2     |    1     |          3          |    5    |    9    |    6    |         |\n",
      "|     6      |      2      |     2     |    1     |          3          |    5    |    6    |    9    |         |\n",
      "|     7      |      1      |     2     |    0     |          1          |    0    |    1    |    8    |         |\n",
      "|     8      |      1      |     2     |    1     |          5          |    4    |    5    |    6    |         |\n",
      "|     9      |      1      |     2     |    1     |          5          |    5    |    6    |    7    |         |\n",
      "|     10     |      1      |     3     |    1     |          5          |    4    |    5    |    6    |    7    |\n",
      "|     11     |      1      |     2     |    1     |          5          |    4    |    5    |    7    |         |\n",
      "|     12     |      1      |     2     |    1     |          5          |    4    |    6    |    7    |         |\n",
      "|     13     |      0      |     2     |    0     |          5          |    0    |    1    |    2    |         |\n",
      "|     14     |      0      |     2     |    0     |          5          |    0    |    1    |    3    |         |\n",
      "|     15     |      0      |     3     |    0     |          5          |    0    |    1    |    2    |    3    |\n",
      "|     16     |      0      |     2     |    0     |          5          |    0    |    2    |    3    |         |\n",
      "|     17     |      0      |     2     |    0     |          5          |    1    |    2    |    3    |         |\n",
      "|     18     |      0      |     2     |    1     |          5          |    4    |    5    |    6    |         |\n",
      "|     19     |      0      |     2     |    1     |          5          |    5    |    6    |    7    |         |\n",
      "|     20     |      0      |     3     |    1     |          5          |    4    |    5    |    6    |    7    |\n",
      "|     21     |      0      |     2     |    1     |          5          |    4    |    5    |    7    |         |\n",
      "|     22     |      0      |     2     |    1     |          5          |    4    |    6    |    7    |         |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+\n"
     ]
    }
   ],
   "source": [
    "twosimplices_preprocessing._view_db('graph')\n",
    "twosimplices_preprocessing._view_db('node_classes')\n",
    "twosimplices_preprocessing._view_db('edge_details')\n",
    "twosimplices_preprocessing._view_db('simplices')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "id": "cc790bb5",
   "metadata": {},
   "outputs": [],
   "source": [
    "src_j = [0,0,1,1,1,5,2,2,3] + [5] + [6,6 ,7,15,15,8] + [7]  + [13,13,10,13,10,11] + [0,14]\n",
    "dst_j = [1,2,2,3,5,3,3,4,4] + [6] + [7,15,15,8,9,9]  + [13] + [10,11,11,12,12,12] + [14,10]\n",
    "jumbo = dgl.heterograph({('paper', 'cites', 'paper'): (src_j, dst_j)})\n",
    "filepath3 = 'jumbo'\n",
    "db3 = 'jumbo.db'\n",
    "jumbo_preprocessing = graph_towers(filepath3,database_name = db3, graph=jumbo,ratio=0.01,bottom_level = 20, max_dimension = 20)\n",
    "jumbo_preprocessing.create_table()\n",
    "#jumbo_preprocessing.empty_table('edge_details')\n",
    "#jumbo_preprocessing.empty_table('graph')\n",
    "#jumbo_preprocessing.empty_table('node_classes')\n",
    "#jumbo_preprocessing.empty_table('simplices')\n",
    "jumbo_preprocessing.create_towers()\n",
    "jumbo_preprocessing.simplicial_search()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "id": "b4890a55",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "| quotient_number | number_of_nodes | number_of_edges | number_of_simplices |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "|        0        |       16        |       25        |          0          |\n",
      "|        1        |       14        |       21        |          0          |\n",
      "|        2        |       13        |       19        |          0          |\n",
      "|        3        |       12        |       16        |          0          |\n",
      "|        4        |       11        |       14        |          0          |\n",
      "|        5        |       10        |       13        |          0          |\n",
      "|        6        |        9        |       12        |          0          |\n",
      "|        7        |        8        |        9        |          0          |\n",
      "|        8        |        7        |        7        |          0          |\n",
      "|        9        |        6        |        5        |          0          |\n",
      "|       10        |        5        |        4        |          0          |\n",
      "|       11        |        3        |        2        |          0          |\n",
      "|       12        |        1        |        1        |          0          |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "| node_class_id | quotient_id | number_of_nodes | globe_id | node_0 | node_1 | node_2 | node_3 | node_4 | node_5 |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "|       0       |      0      |       16        |    0     |   0    |        |        |        |        |        |\n",
      "|       1       |      0      |       16        |    0     |   1    |        |        |        |        |        |\n",
      "|       2       |      0      |       16        |    0     |   2    |        |        |        |        |        |\n",
      "|       3       |      0      |       16        |    0     |   3    |        |        |        |        |        |\n",
      "|       4       |      0      |       16        |    0     |   4    |        |        |        |        |        |\n",
      "|       5       |      0      |       16        |    0     |   5    |        |        |        |        |        |\n",
      "|       6       |      0      |       16        |    1     |   6    |        |        |        |        |        |\n",
      "|       7       |      0      |       16        |    1     |   7    |        |        |        |        |        |\n",
      "|       8       |      0      |       16        |    1     |   8    |        |        |        |        |        |\n",
      "|       9       |      0      |       16        |    1     |   9    |        |        |        |        |        |\n",
      "|      10       |      0      |       16        |    2     |   10   |        |        |        |        |        |\n",
      "|      11       |      0      |       16        |    2     |   11   |        |        |        |        |        |\n",
      "|      12       |      0      |       16        |    2     |   12   |        |        |        |        |        |\n",
      "|      13       |      0      |       16        |    2     |   13   |        |        |        |        |        |\n",
      "|      14       |      0      |       16        |    -1    |   14   |        |        |        |        |        |\n",
      "|      15       |      0      |       16        |    1     |   15   |        |        |        |        |        |\n",
      "|      16       |      1      |       14        |    0     |   0    |   2    |        |        |        |        |\n",
      "|      17       |      2      |       13        |    1     |   15   |   7    |        |        |        |        |\n",
      "|      18       |      3      |       12        |    2     |   11   |   12   |        |        |        |        |\n",
      "|      19       |      4      |       11        |    0     |   3    |   5    |        |        |        |        |\n",
      "|      20       |      5      |       10        |    0     |   16   |   4    |        |        |        |        |\n",
      "|      21       |      6      |        9        |    1     |   17   |   9    |        |        |        |        |\n",
      "|      22       |      7      |        8        |    0     |   19   |   20   |        |        |        |        |\n",
      "|      23       |      8      |        7        |    2     |   10   |   13   |        |        |        |        |\n",
      "|      24       |      9      |        6        |    1     |   8    |   21   |        |        |        |        |\n",
      "|      25       |     10      |        5        |    2     |   18   |   23   |        |        |        |        |\n",
      "|      26       |     11      |        3        |    0     |   1    |   22   |        |        |        |        |\n",
      "|      27       |     12      |        1        |    1     |   24   |   6    |        |        |        |        |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "| e_src | e_dst | edge_changed | to_contract | sampled | contracted | quotient_id | multiplicity | globe_id |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   2   |      1       |      1      |    1    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   5   |   3   |      1       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   5   |   6   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |  15   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |  15   |      1       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|  15   |   8   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|  15   |   9   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |  13   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  10   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  12   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  10   |  12   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  11   |  12   |      1       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|   0   |  14   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|  14   |  10   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|  16   |  16   |      1       |      1      |    1    |     1      |      1      |      1       |    0     |\n",
      "|  16   |   1   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |  16   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   5   |   3   |      1       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|  16   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|  16   |   4   |      1       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   6   |  15   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   7   |  15   |      1       |      1      |    1    |     0      |      1      |      1       |    1     |\n",
      "|  15   |   8   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|  15   |   9   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  11   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  10   |  11   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  12   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  10   |  12   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  11   |  12   |      1       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  17   |  17   |      1       |      1      |    1    |     1      |      2      |      1       |    1     |\n",
      "|  16   |   1   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |  16   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   5   |   3   |      1       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|  16   |   3   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|  16   |   4   |      1       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   6   |  17   |      0       |      1      |    0    |     0      |      2      |      2       |    1     |\n",
      "|  17   |   8   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|  17   |   9   |      1       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  11   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  10   |  11   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  12   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  10   |  12   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  11   |  12   |      1       |      1      |    1    |     0      |      2      |      1       |    2     |\n",
      "|  18   |  18   |      1       |      1      |    1    |     1      |      3      |      1       |    2     |\n",
      "|  16   |   1   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |  16   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   5   |   3   |      1       |      1      |    1    |     0      |      3      |      1       |    0     |\n",
      "|  16   |   3   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|  16   |   4   |      1       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   6   |  17   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|  17   |   8   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|  17   |   9   |      1       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  10   |  18   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  19   |  19   |      1       |      1      |    1    |     1      |      4      |      1       |    0     |\n",
      "|  16   |   1   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   1   |  16   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   1   |  19   |      0       |      1      |    0    |     0      |      4      |      2       |    0     |\n",
      "|  16   |  19   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|  16   |   4   |      1       |      1      |    1    |     0      |      4      |      1       |    0     |\n",
      "|  19   |   4   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   6   |  17   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|  17   |   8   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|  17   |   9   |      1       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  10   |  18   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  20   |  20   |      1       |      1      |    1    |     1      |      5      |      1       |    0     |\n",
      "|  20   |   1   |      0       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|   1   |  20   |      0       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|   1   |  19   |      0       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|  20   |  19   |      1       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|  19   |  20   |      1       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|   6   |  17   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|  17   |   8   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|  17   |   9   |      1       |      1      |    1    |     0      |      5      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  10   |  18   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  21   |  21   |      1       |      1      |    1    |     1      |      6      |      1       |    1     |\n",
      "|  20   |   1   |      0       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|   1   |  20   |      0       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|   1   |  19   |      0       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|  20   |  19   |      1       |      1      |    1    |     0      |      6      |      1       |    0     |\n",
      "|  19   |  20   |      1       |      1      |    1    |     0      |      6      |      1       |    0     |\n",
      "|   6   |  21   |      0       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|  21   |   8   |      1       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|   8   |  21   |      1       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  10   |  18   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  22   |  22   |      1       |      1      |    1    |     1      |      7      |      2       |    0     |\n",
      "|  22   |   1   |      1       |      1      |    0    |     0      |      7      |      1       |    0     |\n",
      "|   1   |  22   |      1       |      1      |    0    |     0      |      7      |      2       |    0     |\n",
      "|   6   |  21   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|  21   |   8   |      1       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|   8   |  21   |      1       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|  13   |  10   |      1       |      1      |    1    |     0      |      7      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  10   |  18   |      0       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  23   |  23   |      1       |      1      |    1    |     1      |      8      |      1       |    2     |\n",
      "|  22   |   1   |      1       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|   1   |  22   |      1       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|   6   |  21   |      0       |      1      |    0    |     0      |      8      |      1       |    1     |\n",
      "|  21   |   8   |      1       |      1      |    1    |     0      |      8      |      1       |    1     |\n",
      "|   8   |  21   |      1       |      1      |    1    |     0      |      8      |      1       |    1     |\n",
      "|  23   |  18   |      1       |      1      |    0    |     0      |      8      |      2       |    2     |\n",
      "|  24   |  24   |      1       |      1      |    1    |     1      |      9      |      2       |    1     |\n",
      "|  22   |   1   |      1       |      1      |    0    |     0      |      9      |      1       |    0     |\n",
      "|   1   |  22   |      1       |      1      |    0    |     0      |      9      |      1       |    0     |\n",
      "|   6   |  24   |      1       |      1      |    0    |     0      |      9      |      1       |    1     |\n",
      "|  23   |  18   |      1       |      1      |    1    |     0      |      9      |      1       |    2     |\n",
      "|  25   |  25   |      0       |      1      |    1    |     1      |     10      |      1       |    2     |\n",
      "|  22   |   1   |      1       |      1      |    1    |     0      |     10      |      1       |    0     |\n",
      "|   1   |  22   |      1       |      1      |    1    |     0      |     10      |      1       |    0     |\n",
      "|   6   |  24   |      1       |      1      |    0    |     0      |     10      |      1       |    1     |\n",
      "|  26   |  26   |      0       |      1      |    1    |     1      |     11      |      2       |    0     |\n",
      "|   6   |  24   |      1       |      1      |    1    |     0      |     11      |      1       |    1     |\n",
      "|  27   |  27   |      0       |      1      |    1    |     1      |     12      |      1       |    1     |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "| simplex_id | quotient_id | dimension | globe_id | number_of_simplices | index_0 | index_1 | index_2 | index_3 | index_4 | index_5 |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "|     1      |      7      |     2     |    2     |          1          |   13    |   10    |   18    |         |         |         |\n",
      "|     2      |      6      |     2     |    0     |          3          |    1    |   19    |   20    |         |         |         |\n",
      "|     3      |      6      |     2     |    0     |          3          |    1    |   20    |   19    |         |         |         |\n",
      "|     4      |      6      |     2     |    0     |          3          |   20    |    1    |   19    |         |         |         |\n",
      "|     5      |      6      |     2     |    2     |          1          |   13    |   10    |   18    |         |         |         |\n",
      "|     6      |      5      |     2     |    0     |          3          |    1    |   19    |   20    |         |         |         |\n",
      "|     7      |      5      |     2     |    0     |          3          |    1    |   20    |   19    |         |         |         |\n",
      "|     8      |      5      |     2     |    0     |          3          |   20    |    1    |   19    |         |         |         |\n",
      "|     9      |      5      |     2     |    1     |          1          |   17    |    8    |    9    |         |         |         |\n",
      "|     10     |      5      |     2     |    2     |          1          |   13    |   10    |   18    |         |         |         |\n",
      "|     11     |      4      |     2     |    0     |          3          |    1    |   16    |   19    |         |         |         |\n",
      "|     12     |      4      |     2     |    0     |          3          |   16    |    1    |   19    |         |         |         |\n",
      "|     13     |      4      |     2     |    0     |          3          |   16    |   19    |    4    |         |         |         |\n",
      "|     14     |      4      |     2     |    1     |          1          |   17    |    8    |    9    |         |         |         |\n",
      "|     15     |      4      |     2     |    2     |          1          |   13    |   10    |   18    |         |         |         |\n",
      "|     16     |      3      |     2     |    0     |          4          |    1    |   16    |    3    |         |         |         |\n",
      "|     17     |      3      |     2     |    0     |          4          |   16    |    1    |    3    |         |         |         |\n",
      "|     18     |      3      |     2     |    0     |          4          |   16    |    3    |    4    |         |         |         |\n",
      "|     19     |      3      |     2     |    0     |          4          |    1    |    5    |    3    |         |         |         |\n",
      "|     20     |      3      |     2     |    1     |          1          |   17    |    8    |    9    |         |         |         |\n",
      "|     21     |      3      |     2     |    2     |          1          |   13    |   10    |   18    |         |         |         |\n",
      "|     22     |      2      |     2     |    0     |          4          |    1    |   16    |    3    |         |         |         |\n",
      "|     23     |      2      |     2     |    0     |          4          |   16    |    1    |    3    |         |         |         |\n",
      "|     24     |      2      |     2     |    0     |          4          |   16    |    3    |    4    |         |         |         |\n",
      "|     25     |      2      |     2     |    0     |          4          |    1    |    5    |    3    |         |         |         |\n",
      "|     26     |      2      |     2     |    1     |          1          |   17    |    8    |    9    |         |         |         |\n",
      "|     27     |      2      |     2     |    2     |          5          |   13    |   10    |   11    |         |         |         |\n",
      "|     28     |      2      |     2     |    2     |          5          |   13    |   10    |   12    |         |         |         |\n",
      "|     29     |      2      |     3     |    2     |          5          |   13    |   10    |   11    |   12    |         |         |\n",
      "|     30     |      2      |     2     |    2     |          5          |   13    |   11    |   12    |         |         |         |\n",
      "|     31     |      2      |     2     |    2     |          5          |   10    |   11    |   12    |         |         |         |\n",
      "|     32     |      1      |     2     |    0     |          4          |    1    |   16    |    3    |         |         |         |\n",
      "|     33     |      1      |     2     |    0     |          4          |   16    |    1    |    3    |         |         |         |\n",
      "|     34     |      1      |     2     |    0     |          4          |   16    |    3    |    4    |         |         |         |\n",
      "|     35     |      1      |     2     |    0     |          4          |    1    |    5    |    3    |         |         |         |\n",
      "|     36     |      1      |     2     |    1     |          2          |   15    |    8    |    9    |         |         |         |\n",
      "|     37     |      1      |     2     |    1     |          2          |    6    |    7    |   15    |         |         |         |\n",
      "|     38     |      1      |     2     |    2     |          5          |   13    |   10    |   11    |         |         |         |\n",
      "|     39     |      1      |     2     |    2     |          5          |   13    |   10    |   12    |         |         |         |\n",
      "|     40     |      1      |     3     |    2     |          5          |   13    |   10    |   11    |   12    |         |         |\n",
      "|     41     |      1      |     2     |    2     |          5          |   13    |   11    |   12    |         |         |         |\n",
      "|     42     |      1      |     2     |    2     |          5          |   10    |   11    |   12    |         |         |         |\n",
      "|     43     |      0      |     2     |    0     |          4          |    1    |    2    |    3    |         |         |         |\n",
      "|     44     |      0      |     2     |    0     |          4          |    2    |    3    |    4    |         |         |         |\n",
      "|     45     |      0      |     2     |    0     |          4          |    1    |    5    |    3    |         |         |         |\n",
      "|     46     |      0      |     2     |    0     |          4          |    0    |    1    |    2    |         |         |         |\n",
      "|     47     |      0      |     2     |    1     |          2          |   15    |    8    |    9    |         |         |         |\n",
      "|     48     |      0      |     2     |    1     |          2          |    6    |    7    |   15    |         |         |         |\n",
      "|     49     |      0      |     2     |    2     |          5          |   13    |   10    |   11    |         |         |         |\n",
      "|     50     |      0      |     2     |    2     |          5          |   13    |   10    |   12    |         |         |         |\n",
      "|     51     |      0      |     3     |    2     |          5          |   13    |   10    |   11    |   12    |         |         |\n",
      "|     52     |      0      |     2     |    2     |          5          |   13    |   11    |   12    |         |         |         |\n",
      "|     53     |      0      |     2     |    2     |          5          |   10    |   11    |   12    |         |         |         |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n"
     ]
    }
   ],
   "source": [
    "jumbo_preprocessing._view_db('graph')\n",
    "jumbo_preprocessing._view_db('node_classes')\n",
    "jumbo_preprocessing._view_db('edge_details')\n",
    "jumbo_preprocessing._view_db('simplices')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "id": "5cb5a406",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "         15812 function calls (15797 primitive calls) in 0.447 seconds\n",
      "\n",
      "   Ordered by: standard name\n",
      "\n",
      "   ncalls  tottime  percall  cumtime  percall filename:lineno(function)\n",
      "       32    0.005    0.000    0.028    0.001 279130135.py:101(_connect_db)\n",
      "        1    0.007    0.007    0.049    0.049 279130135.py:179(initial_db_fill)\n",
      "       16    0.002    0.000    0.211    0.013 279130135.py:245(sampling)\n",
      "       16    0.000    0.000    0.000    0.000 279130135.py:260(<dictcomp>)\n",
      "       15    0.002    0.000    0.003    0.000 279130135.py:317(make_quotient)\n",
      "       15    0.000    0.000    0.000    0.000 279130135.py:323(<dictcomp>)\n",
      "       15    0.003    0.000    0.183    0.012 279130135.py:363(db_fill)\n",
      "       15    0.000    0.000    0.000    0.000 279130135.py:425(<listcomp>)\n",
      "        1    0.001    0.001    0.447    0.447 279130135.py:441(create_towers)\n",
      "       16    0.001    0.000    0.001    0.000 279130135.py:620(relation)\n",
      "    75/60    0.000    0.000    0.000    0.000 279130135.py:624(parent)\n",
      "       16    0.000    0.000    0.000    0.000 279130135.py:646(<listcomp>)\n",
      "       32    0.001    0.000    0.247    0.008 279130135.py:97(_close_db)\n",
      "      330    0.000    0.000    0.001    0.000 <frozen abc>:117(__instancecheck__)\n",
      "      217    0.000    0.000    0.000    0.000 <frozen abc>:121(__subclasscheck__)\n",
      "        1    0.000    0.000    0.447    0.447 <string>:1(<module>)\n",
      "      147    0.000    0.000    0.008    0.000 __init__.py:145(from_dgl_nd)\n",
      "      294    0.001    0.000    0.001    0.000 __init__.py:511(cast)\n",
      "       15    0.000    0.000    0.001    0.000 __init__.py:587(__init__)\n",
      "       15    0.000    0.000    0.000    0.000 __init__.py:660(update)\n",
      "       19    0.000    0.000    0.000    0.000 _tensor.py:1002(__hash__)\n",
      "       85    0.001    0.000    0.002    0.000 _tensor.py:964(__len__)\n",
      "        2    0.000    0.000    0.000    0.000 _tensor.py:980(__iter__)\n",
      "       21    0.000    0.000    0.000    0.000 base.py:23(<lambda>)\n",
      "      798    0.000    0.000    0.000    0.000 base.py:58(check_call)\n",
      "       49    0.000    0.000    0.000    0.000 base.py:73(c_str)\n",
      "      147    0.000    0.000    0.001    0.000 dlpack.py:49(from_dlpack)\n",
      "      357    0.004    0.000    0.006    0.000 function.py:103(_make_dgl_args)\n",
      "       49    0.000    0.000    0.000    0.000 function.py:184(__init__)\n",
      "       49    0.000    0.000    0.000    0.000 function.py:198(__del__)\n",
      "      357    0.004    0.000    0.013    0.000 function.py:202(__call__)\n",
      "       49    0.000    0.000    0.000    0.000 function.py:258(_handle_return_func)\n",
      "      147    0.000    0.000    0.001    0.000 function.py:270(<lambda>)\n",
      "       32    0.001    0.000    0.001    0.000 functools.py:35(update_wrapper)\n",
      "       32    0.000    0.000    0.000    0.000 functools.py:479(lru_cache)\n",
      "       32    0.000    0.000    0.002    0.000 functools.py:518(decorating_function)\n",
      "       49    0.000    0.000    0.001    0.000 graph_index.py:174(num_edges)\n",
      "       49    0.000    0.000    0.000    0.000 graph_index.py:39(__new__)\n",
      "       21    0.000    0.000    0.000    0.000 heterograph.py:1297(get_ntype_id)\n",
      "       49    0.000    0.000    0.004    0.000 heterograph.py:1389(get_etype_id)\n",
      "       21    0.000    0.000    0.000    0.000 heterograph.py:1775(nodes)\n",
      "       49    0.000    0.000    0.000    0.000 heterograph.py:2154(edges)\n",
      "       21    0.000    0.000    0.001    0.000 heterograph.py:2831(idtype)\n",
      "       49    0.000    0.000    0.023    0.000 heterograph.py:3519(all_edges)\n",
      "       21    0.000    0.000    0.001    0.000 heterograph.py:5629(device)\n",
      "       42    0.000    0.000    0.000    0.000 heterograph.py:959(is_unibipartite)\n",
      "       21    0.000    0.000    0.001    0.000 heterograph_index.py:178(dtype)\n",
      "       21    0.000    0.000    0.001    0.000 heterograph_index.py:189(ctx)\n",
      "       21    0.000    0.000    0.001    0.000 heterograph_index.py:382(num_nodes)\n",
      "       49    0.001    0.000    0.018    0.000 heterograph_index.py:666(edges)\n",
      "       49    0.000    0.000    0.002    0.000 heterograph_index.py:74(metagraph)\n",
      "       49    0.000    0.000    0.004    0.000 heterograph_index.py:93(number_of_etypes)\n",
      "      147    0.000    0.000    0.001    0.000 ndarray.py:113(_make_array)\n",
      "      441    0.002    0.000    0.003    0.000 ndarray.py:206(shape)\n",
      "      882    0.001    0.000    0.001    0.000 ndarray.py:209(<genexpr>)\n",
      "      147    0.001    0.000    0.001    0.000 ndarray.py:48(_dlpack_deleter)\n",
      "      147    0.000    0.000    0.000    0.000 ndarray.py:70(__init__)\n",
      "      147    0.000    0.000    0.001    0.000 ndarray.py:81(__del__)\n",
      "      147    0.001    0.000    0.002    0.000 ndarray.py:89(to_dlpack)\n",
      "       49    0.000    0.000    0.001    0.000 object.py:28(_return_object)\n",
      "       49    0.000    0.000    0.000    0.000 object.py:55(__del__)\n",
      "       21    0.000    0.000    0.000    0.000 runtime_ctypes.py:151(__new__)\n",
      "       21    0.000    0.000    0.000    0.000 runtime_ctypes.py:163(__init__)\n",
      "       21    0.000    0.000    0.000    0.000 tensor.py:115(to_backend_ctx)\n",
      "       21    0.000    0.000    0.000    0.000 tensor.py:406(arange)\n",
      "      147    0.002    0.000    0.008    0.000 tensor.py:465(zerocopy_from_dgl_ndarray)\n",
      "       70    0.000    0.000    0.000    0.000 types.py:70(<lambda>)\n",
      "       21    0.000    0.000    0.000    0.000 types.py:74(<lambda>)\n",
      "       21    0.000    0.000    0.000    0.000 types.py:76(<lambda>)\n",
      "       49    0.000    0.000    0.000    0.000 view.py:147(__init__)\n",
      "       49    0.000    0.000    0.023    0.000 view.py:177(__call__)\n",
      "       21    0.000    0.000    0.000    0.000 view.py:20(__init__)\n",
      "       21    0.000    0.000    0.003    0.000 view.py:46(__call__)\n",
      "       49    0.000    0.000    0.000    0.000 {built-in method __new__ of type object at 0x00007FFB9675A4B0}\n",
      "      330    0.001    0.000    0.001    0.000 {built-in method _abc._abc_instancecheck}\n",
      "      217    0.000    0.000    0.000    0.000 {built-in method _abc._abc_subclasscheck}\n",
      "       15    0.000    0.000    0.000    0.000 {built-in method _collections._count_elements}\n",
      "      910    0.000    0.000    0.000    0.000 {built-in method _ctypes.byref}\n",
      "       32    0.021    0.001    0.023    0.001 {built-in method _sqlite3.connect}\n",
      "        1    0.000    0.000    0.447    0.447 {built-in method builtins.exec}\n",
      "      245    0.001    0.000    0.001    0.000 {built-in method builtins.getattr}\n",
      "      196    0.000    0.000    0.000    0.000 {built-in method builtins.hasattr}\n",
      "      944    0.000    0.000    0.000    0.000 {built-in method builtins.id}\n",
      "     1930    0.001    0.000    0.002    0.000 {built-in method builtins.isinstance}\n",
      "       32    0.000    0.000    0.000    0.000 {built-in method builtins.iter}\n",
      "      778    0.001    0.000    0.002    0.000 {built-in method builtins.len}\n",
      "      147    0.000    0.000    0.000    0.000 {built-in method builtins.min}\n",
      "       30    0.000    0.000    0.000    0.000 {built-in method builtins.next}\n",
      "       64    0.000    0.000    0.000    0.000 {built-in method builtins.setattr}\n",
      "       16    0.000    0.000    0.000    0.000 {built-in method math.ceil}\n",
      "      147    0.001    0.000    0.001    0.000 {built-in method torch._C._from_dlpack}\n",
      "       87    0.001    0.000    0.001    0.000 {built-in method torch._C._get_tracing_state}\n",
      "       85    0.000    0.000    0.000    0.000 {built-in method torch._C._has_torch_function_unary}\n",
      "       51    0.001    0.000    0.001    0.000 {built-in method torch.all}\n",
      "       51    0.001    0.000    0.001    0.000 {built-in method torch.any}\n",
      "       21    0.000    0.000    0.000    0.000 {built-in method torch.arange}\n",
      "       60    0.000    0.000    0.000    0.000 {method 'add' of 'set' objects}\n",
      "      274    0.000    0.000    0.000    0.000 {method 'append' of 'list' objects}\n",
      "       32    0.007    0.000    0.007    0.000 {method 'close' of 'sqlite3.Connection' objects}\n",
      "       32    0.239    0.007    0.239    0.007 {method 'commit' of 'sqlite3.Connection' objects}\n",
      "       32    0.000    0.000    0.000    0.000 {method 'cursor' of 'sqlite3.Connection' objects}\n",
      "       21    0.000    0.000    0.000    0.000 {method 'decode' of 'bytes' objects}\n",
      "       87    0.000    0.000    0.000    0.000 {method 'dim' of 'torch._C._TensorBase' objects}\n",
      "        1    0.000    0.000    0.000    0.000 {method 'disable' of '_lsprof.Profiler' objects}\n",
      "       49    0.000    0.000    0.000    0.000 {method 'encode' of 'str' objects}\n",
      "      161    0.113    0.001    0.113    0.001 {method 'execute' of 'sqlite3.Cursor' objects}\n",
      "       30    0.010    0.000    0.010    0.000 {method 'executemany' of 'sqlite3.Cursor' objects}\n",
      "       47    0.001    0.000    0.001    0.000 {method 'fetchall' of 'sqlite3.Cursor' objects}\n",
      "       46    0.000    0.000    0.000    0.000 {method 'format' of 'str' objects}\n",
      "     1234    0.000    0.000    0.000    0.000 {method 'get' of 'dict' objects}\n",
      "       45    0.000    0.000    0.000    0.000 {method 'items' of 'dict' objects}\n",
      "      122    0.000    0.000    0.000    0.000 {method 'join' of 'str' objects}\n",
      "       62    0.000    0.000    0.000    0.000 {method 'keys' of 'dict' objects}\n",
      "        2    0.000    0.000    0.000    0.000 {method 'unbind' of 'torch._C._TensorBase' objects}\n",
      "      289    0.000    0.000    0.000    0.000 {method 'union' of 'set' objects}\n",
      "       47    0.000    0.000    0.000    0.000 {method 'update' of 'dict' objects}\n",
      "       32    0.000    0.000    0.000    0.000 {method 'values' of 'dict' objects}\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "multiple_5_simplices_src = [0,0,0,0,0,1,1,1,1,2,2,2,3,3,4] + [3] + [6,6,6,6, 6, 7,7,7, 7, 8, 8, 8, 9, 9,10] + [11,12] + [13,13,13,13,13,14,14,14,14,15,15,15,16,16,17]\n",
    "multiple_5_simplices_dst = [1,2,3,4,5,2,3,4,5,3,4,5,4,5,5] + [6] + [7,8,9,10,11,8,9,10,11,9,10,11,10,11,11] + [12,13] + [14,15,16,17,18,15,16,17,18,16,17,18,17,18,18]\n",
    "multiple_5_simplices = dgl.heterograph({('paper', 'cites', 'paper'): (multiple_5_simplices_src, multiple_5_simplices_dst)})\n",
    "filepath = 'multiple_5_simplices'\n",
    "db4 = 'multiple_5_simplices.db'\n",
    "mulitpl5_preprocessing = graph_towers(filepath,database_name = db4, graph=multiple_5_simplices,ratio=0.01,bottom_level = 20, max_dimension = 20)\n",
    "mulitpl5_preprocessing.create_table()\n",
    "#mulitpl5_preprocessing.empty_table('edge_details')\n",
    "#mulitpl5_preprocessing.empty_table('graph')\n",
    "#mulitpl5_preprocessing.empty_table('node_classes')\n",
    "#mulitpl5_preprocessing.empty_table('simplices')\n",
    "mulitpl5_preprocessing.create_towers()\n",
    "mulitpl5_preprocessing.simplicial_search()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "id": "14efe77f",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "| quotient_number | number_of_nodes | number_of_edges | number_of_simplices |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "|        0        |       19        |       48        |          0          |\n",
      "|        1        |       17        |       41        |          0          |\n",
      "|        2        |       16        |       37        |          0          |\n",
      "|        3        |       15        |       32        |          0          |\n",
      "|        4        |       14        |       29        |          0          |\n",
      "|        5        |       13        |       25        |          0          |\n",
      "|        6        |       12        |       22        |          0          |\n",
      "|        7        |       11        |       19        |          0          |\n",
      "|        8        |       10        |       15        |          0          |\n",
      "|        9        |        9        |       11        |          0          |\n",
      "|       10        |        8        |        8        |          0          |\n",
      "|       11        |        7        |        7        |          0          |\n",
      "|       12        |        6        |        5        |          0          |\n",
      "|       13        |        4        |        4        |          0          |\n",
      "|       14        |        2        |        3        |          0          |\n",
      "|       15        |        1        |        1        |          0          |\n",
      "+-----------------+-----------------+-----------------+---------------------+\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "| node_class_id | quotient_id | number_of_nodes | globe_id | node_0 | node_1 | node_2 | node_3 | node_4 | node_5 |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "|       0       |      0      |       19        |    0     |   0    |        |        |        |        |        |\n",
      "|       1       |      0      |       19        |    0     |   1    |        |        |        |        |        |\n",
      "|       2       |      0      |       19        |    0     |   2    |        |        |        |        |        |\n",
      "|       3       |      0      |       19        |    0     |   3    |        |        |        |        |        |\n",
      "|       4       |      0      |       19        |    0     |   4    |        |        |        |        |        |\n",
      "|       5       |      0      |       19        |    0     |   5    |        |        |        |        |        |\n",
      "|       6       |      0      |       19        |    1     |   6    |        |        |        |        |        |\n",
      "|       7       |      0      |       19        |    1     |   7    |        |        |        |        |        |\n",
      "|       8       |      0      |       19        |    1     |   8    |        |        |        |        |        |\n",
      "|       9       |      0      |       19        |    1     |   9    |        |        |        |        |        |\n",
      "|      10       |      0      |       19        |    1     |   10   |        |        |        |        |        |\n",
      "|      11       |      0      |       19        |    1     |   11   |        |        |        |        |        |\n",
      "|      12       |      0      |       19        |    -1    |   12   |        |        |        |        |        |\n",
      "|      13       |      0      |       19        |    2     |   13   |        |        |        |        |        |\n",
      "|      14       |      0      |       19        |    2     |   14   |        |        |        |        |        |\n",
      "|      15       |      0      |       19        |    2     |   15   |        |        |        |        |        |\n",
      "|      16       |      0      |       19        |    2     |   16   |        |        |        |        |        |\n",
      "|      17       |      0      |       19        |    2     |   17   |        |        |        |        |        |\n",
      "|      18       |      0      |       19        |    2     |   18   |        |        |        |        |        |\n",
      "|      19       |      1      |       17        |    1     |   10   |   11   |        |        |        |        |\n",
      "|      20       |      2      |       16        |    0     |   0    |   2    |        |        |        |        |\n",
      "|      21       |      3      |       15        |    2     |   13   |   14   |        |        |        |        |\n",
      "|      22       |      4      |       14        |    0     |   1    |   4    |        |        |        |        |\n",
      "|      23       |      5      |       13        |    0     |   20   |   22   |        |        |        |        |\n",
      "|      24       |      6      |       12        |    1     |   8    |   6    |        |        |        |        |\n",
      "|      25       |      7      |       11        |    2     |   16   |   18   |        |        |        |        |\n",
      "|      26       |      8      |       10        |    1     |   24   |   7    |        |        |        |        |\n",
      "|      27       |      9      |        9        |    2     |   17   |   25   |        |        |        |        |\n",
      "|      28       |     10      |        8        |    0     |   3    |   23   |        |        |        |        |\n",
      "|      29       |     11      |        7        |    1     |   26   |   19   |        |        |        |        |\n",
      "|      30       |     12      |        6        |    1     |   9    |   29   |        |        |        |        |\n",
      "|      31       |     13      |        4        |    0     |   28   |   5    |        |        |        |        |\n",
      "|      32       |     14      |        2        |    2     |   27   |   21   |        |        |        |        |\n",
      "|      33       |     15      |        1        |    2     |   32   |   15   |        |        |        |        |\n",
      "+---------------+-------------+-----------------+----------+--------+--------+--------+--------+--------+--------+\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "| e_src | e_dst | edge_changed | to_contract | sampled | contracted | quotient_id | multiplicity | globe_id |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   2   |      1       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   0   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   2   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   4   |      1       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   3   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      0      |      1       |    0     |\n",
      "|   3   |   6   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |  10   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   6   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |  10   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   7   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   8   |  10   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   8   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   9   |  10   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|   9   |  11   |      0       |      1      |    0    |     0      |      0      |      1       |    1     |\n",
      "|  10   |  11   |      1       |      1      |    1    |     0      |      0      |      1       |    1     |\n",
      "|  11   |  12   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|  12   |  13   |      0       |      0      |    0    |     0      |      0      |      1       |    -1    |\n",
      "|  13   |  14   |      1       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  15   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  16   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  17   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  14   |  15   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  14   |  16   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  14   |  17   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  14   |  18   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      0      |      1       |    2     |\n",
      "|  19   |  19   |      1       |      1      |    1    |     1      |      1      |      1       |    1     |\n",
      "|   0   |   1   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   2   |      1       |      1      |    1    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   4   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   0   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   2   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   4   |      1       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   2   |   3   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   2   |   4   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   2   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      1      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   6   |  19   |      0       |      1      |    0    |     0      |      1      |      2       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      1      |      2       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      1      |      1       |    1     |\n",
      "|   8   |  19   |      0       |      1      |    0    |     0      |      1      |      2       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      1      |      2       |    1     |\n",
      "|  13   |  14   |      1       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  15   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  16   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  17   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  14   |  15   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  14   |  16   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  14   |  17   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  14   |  18   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      1      |      1       |    2     |\n",
      "|  20   |  20   |      1       |      1      |    1    |     1      |      2      |      1       |    0     |\n",
      "|  20   |   1   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|  20   |   3   |      0       |      1      |    0    |     0      |      2      |      2       |    0     |\n",
      "|  20   |   4   |      0       |      1      |    0    |     0      |      2      |      2       |    0     |\n",
      "|  20   |   5   |      0       |      1      |    0    |     0      |      2      |      2       |    0     |\n",
      "|   1   |  20   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   4   |      1       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      2      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   6   |  19   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   8   |  19   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      2      |      1       |    1     |\n",
      "|  13   |  14   |      1       |      1      |    1    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  15   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  16   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  17   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  13   |  18   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  14   |  15   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  14   |  16   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  14   |  17   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  14   |  18   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      2      |      1       |    2     |\n",
      "|  21   |  21   |      1       |      1      |    1    |     1      |      3      |      1       |    2     |\n",
      "|  20   |   1   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|  20   |   3   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|  20   |   4   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|  20   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |  20   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   3   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   4   |      1       |      1      |    1    |     0      |      3      |      1       |    0     |\n",
      "|   1   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   3   |   4   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   4   |   5   |      0       |      1      |    0    |     0      |      3      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   6   |  19   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   8   |  19   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      3      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  21   |  16   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  21   |  18   |      0       |      1      |    0    |     0      |      3      |      2       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      3      |      1       |    2     |\n",
      "|  22   |  22   |      1       |      1      |    1    |     1      |      4      |      1       |    0     |\n",
      "|  20   |  22   |      1       |      1      |    1    |     0      |      4      |      2       |    0     |\n",
      "|  20   |   3   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|  20   |   5   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|  22   |  20   |      1       |      1      |    1    |     0      |      4      |      1       |    0     |\n",
      "|  22   |   3   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|  22   |   5   |      0       |      1      |    0    |     0      |      4      |      2       |    0     |\n",
      "|   3   |  22   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      4      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   6   |  19   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   8   |  19   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      4      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  21   |  16   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  21   |  18   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      4      |      1       |    2     |\n",
      "|  23   |  23   |      1       |      1      |    1    |     1      |      5      |      2       |    0     |\n",
      "|  23   |   3   |      1       |      1      |    0    |     0      |      5      |      2       |    0     |\n",
      "|  23   |   5   |      0       |      1      |    0    |     0      |      5      |      2       |    0     |\n",
      "|   3   |  23   |      1       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      5      |      1       |    0     |\n",
      "|   6   |   7   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   6   |   8   |      1       |      1      |    1    |     0      |      5      |      1       |    1     |\n",
      "|   6   |   9   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   6   |  19   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   7   |   8   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   8   |   9   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   8   |  19   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      5      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  21   |  16   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  21   |  18   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      5      |      1       |    2     |\n",
      "|  24   |  24   |      1       |      1      |    1    |     1      |      6      |      1       |    1     |\n",
      "|  23   |   3   |      1       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|  23   |   5   |      0       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|   3   |  23   |      1       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      6      |      1       |    0     |\n",
      "|  24   |   7   |      1       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|  24   |   9   |      0       |      1      |    0    |     0      |      6      |      2       |    1     |\n",
      "|  24   |  19   |      0       |      1      |    0    |     0      |      6      |      2       |    1     |\n",
      "|   7   |  24   |      1       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      6      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  21   |  16   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  21   |  18   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  15   |  16   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  15   |  18   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  16   |  17   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  16   |  18   |      1       |      1      |    1    |     0      |      6      |      1       |    2     |\n",
      "|  17   |  18   |      0       |      1      |    0    |     0      |      6      |      1       |    2     |\n",
      "|  25   |  25   |      1       |      1      |    1    |     1      |      7      |      1       |    2     |\n",
      "|  23   |   3   |      1       |      1      |    0    |     0      |      7      |      1       |    0     |\n",
      "|  23   |   5   |      0       |      1      |    0    |     0      |      7      |      1       |    0     |\n",
      "|   3   |  23   |      1       |      1      |    0    |     0      |      7      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      7      |      1       |    0     |\n",
      "|  24   |   7   |      1       |      1      |    1    |     0      |      7      |      1       |    1     |\n",
      "|  24   |   9   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|  24   |  19   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|   7   |  24   |      1       |      1      |    1    |     0      |      7      |      1       |    1     |\n",
      "|   7   |   9   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|   7   |  19   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      7      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  21   |  25   |      0       |      1      |    0    |     0      |      7      |      2       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  15   |  25   |      0       |      1      |    0    |     0      |      7      |      2       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  25   |  17   |      1       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  17   |  25   |      1       |      1      |    0    |     0      |      7      |      1       |    2     |\n",
      "|  26   |  26   |      1       |      1      |    1    |     1      |      8      |      2       |    1     |\n",
      "|  23   |   3   |      1       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|  23   |   5   |      0       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|   3   |  23   |      1       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      8      |      1       |    0     |\n",
      "|  26   |   9   |      0       |      1      |    0    |     0      |      8      |      2       |    1     |\n",
      "|  26   |  19   |      1       |      1      |    0    |     0      |      8      |      2       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      8      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      8      |      1       |    2     |\n",
      "|  21   |  25   |      0       |      1      |    0    |     0      |      8      |      1       |    2     |\n",
      "|  21   |  17   |      0       |      1      |    0    |     0      |      8      |      1       |    2     |\n",
      "|  15   |  25   |      0       |      1      |    0    |     0      |      8      |      1       |    2     |\n",
      "|  15   |  17   |      0       |      1      |    0    |     0      |      8      |      1       |    2     |\n",
      "|  25   |  17   |      1       |      1      |    1    |     0      |      8      |      1       |    2     |\n",
      "|  17   |  25   |      1       |      1      |    1    |     0      |      8      |      1       |    2     |\n",
      "|  27   |  27   |      1       |      1      |    1    |     1      |      9      |      2       |    2     |\n",
      "|  23   |   3   |      1       |      1      |    1    |     0      |      9      |      1       |    0     |\n",
      "|  23   |   5   |      0       |      1      |    0    |     0      |      9      |      1       |    0     |\n",
      "|   3   |  23   |      1       |      1      |    1    |     0      |      9      |      1       |    0     |\n",
      "|   3   |   5   |      0       |      1      |    0    |     0      |      9      |      1       |    0     |\n",
      "|  26   |   9   |      0       |      1      |    0    |     0      |      9      |      1       |    1     |\n",
      "|  26   |  19   |      1       |      1      |    0    |     0      |      9      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |      9      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |      9      |      1       |    2     |\n",
      "|  21   |  27   |      1       |      1      |    0    |     0      |      9      |      2       |    2     |\n",
      "|  15   |  27   |      0       |      1      |    0    |     0      |      9      |      2       |    2     |\n",
      "|  28   |  28   |      1       |      1      |    1    |     1      |     10      |      2       |    0     |\n",
      "|  28   |   5   |      1       |      1      |    0    |     0      |     10      |      2       |    0     |\n",
      "|  26   |   9   |      0       |      1      |    0    |     0      |     10      |      1       |    1     |\n",
      "|  26   |  19   |      1       |      1      |    1    |     0      |     10      |      1       |    1     |\n",
      "|   9   |  19   |      0       |      1      |    0    |     0      |     10      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |     10      |      1       |    2     |\n",
      "|  21   |  27   |      1       |      1      |    0    |     0      |     10      |      1       |    2     |\n",
      "|  15   |  27   |      0       |      1      |    0    |     0      |     10      |      1       |    2     |\n",
      "|  29   |  29   |      1       |      1      |    1    |     1      |     11      |      1       |    1     |\n",
      "|  28   |   5   |      1       |      1      |    0    |     0      |     11      |      1       |    0     |\n",
      "|  29   |   9   |      1       |      1      |    1    |     0      |     11      |      1       |    1     |\n",
      "|   9   |  29   |      1       |      1      |    1    |     0      |     11      |      1       |    1     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |     11      |      1       |    2     |\n",
      "|  21   |  27   |      1       |      1      |    0    |     0      |     11      |      1       |    2     |\n",
      "|  15   |  27   |      0       |      1      |    0    |     0      |     11      |      1       |    2     |\n",
      "|  30   |  30   |      0       |      1      |    1    |     1      |     12      |      2       |    1     |\n",
      "|  28   |   5   |      1       |      1      |    1    |     0      |     12      |      1       |    0     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |     12      |      1       |    2     |\n",
      "|  21   |  27   |      1       |      1      |    0    |     0      |     12      |      1       |    2     |\n",
      "|  15   |  27   |      0       |      1      |    0    |     0      |     12      |      1       |    2     |\n",
      "|  31   |  31   |      0       |      1      |    1    |     1      |     13      |      1       |    0     |\n",
      "|  21   |  15   |      0       |      1      |    0    |     0      |     13      |      1       |    2     |\n",
      "|  21   |  27   |      1       |      1      |    1    |     0      |     13      |      1       |    2     |\n",
      "|  15   |  27   |      0       |      1      |    0    |     0      |     13      |      1       |    2     |\n",
      "|  32   |  32   |      1       |      1      |    1    |     1      |     14      |      1       |    2     |\n",
      "|  32   |  15   |      1       |      1      |    1    |     0      |     14      |      1       |    2     |\n",
      "|  15   |  32   |      1       |      1      |    1    |     0      |     14      |      1       |    2     |\n",
      "|  33   |  33   |      0       |      1      |    1    |     1      |     15      |      2       |    2     |\n",
      "+-------+-------+--------------+-------------+---------+------------+-------------+--------------+----------+\n"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "| simplex_id | quotient_id | dimension | globe_id | number_of_simplices | index_0 | index_1 | index_2 | index_3 | index_4 | index_5 |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n",
      "|     1      |     13      |     2     |    2     |          1          |   21    |   15    |   27    |         |         |         |\n",
      "|     2      |     12      |     2     |    2     |          1          |   21    |   15    |   27    |         |         |         |\n",
      "|     3      |     11      |     2     |    2     |          1          |   21    |   15    |   27    |         |         |         |\n",
      "|     4      |     10      |     2     |    1     |          1          |   26    |    9    |   19    |         |         |         |\n",
      "|     5      |     10      |     2     |    2     |          1          |   21    |   15    |   27    |         |         |         |\n",
      "|     6      |      9      |     2     |    0     |          2          |    3    |   23    |    5    |         |         |         |\n",
      "|     7      |      9      |     2     |    0     |          2          |   23    |    3    |    5    |         |         |         |\n",
      "|     8      |      9      |     2     |    1     |          1          |   26    |    9    |   19    |         |         |         |\n",
      "|     9      |      9      |     2     |    2     |          1          |   21    |   15    |   27    |         |         |         |\n",
      "|     10     |      8      |     2     |    0     |          2          |    3    |   23    |    5    |         |         |         |\n",
      "|     11     |      8      |     2     |    0     |          2          |   23    |    3    |    5    |         |         |         |\n",
      "|     12     |      8      |     2     |    1     |          1          |   26    |    9    |   19    |         |         |         |\n",
      "|     13     |      8      |     2     |    2     |          8          |   21    |   15    |   17    |         |         |         |\n",
      "|     14     |      8      |     2     |    2     |          8          |   21    |   15    |   25    |         |         |         |\n",
      "|     15     |      8      |     3     |    2     |          8          |   21    |   15    |   17    |   25    |         |         |\n",
      "|     16     |      8      |     3     |    2     |          8          |   21    |   15    |   25    |   17    |         |         |\n",
      "|     17     |      8      |     2     |    2     |          8          |   21    |   17    |   25    |         |         |         |\n",
      "|     18     |      8      |     2     |    2     |          8          |   21    |   25    |   17    |         |         |         |\n",
      "|     19     |      8      |     2     |    2     |          8          |   15    |   17    |   25    |         |         |         |\n",
      "|     20     |      8      |     2     |    2     |          8          |   15    |   25    |   17    |         |         |         |\n",
      "|     21     |      7      |     2     |    0     |          2          |    3    |   23    |    5    |         |         |         |\n",
      "|     22     |      7      |     2     |    0     |          2          |   23    |    3    |    5    |         |         |         |\n",
      "|     23     |      7      |     2     |    1     |          8          |   24    |    9    |   19    |         |         |         |\n",
      "|     24     |      7      |     2     |    1     |          8          |    7    |    9    |   19    |         |         |         |\n",
      "|     25     |      7      |     3     |    1     |          8          |   24    |    7    |    9    |   19    |         |         |\n",
      "|     26     |      7      |     3     |    1     |          8          |    7    |   24    |    9    |   19    |         |         |\n",
      "|     27     |      7      |     2     |    1     |          8          |   24    |    7    |    9    |         |         |         |\n",
      "|     28     |      7      |     2     |    1     |          8          |    7    |   24    |    9    |         |         |         |\n",
      "|     29     |      7      |     2     |    1     |          8          |   24    |    7    |   19    |         |         |         |\n",
      "|     30     |      7      |     2     |    1     |          8          |    7    |   24    |   19    |         |         |         |\n",
      "|     31     |      7      |     2     |    2     |          8          |   21    |   15    |   17    |         |         |         |\n",
      "|     32     |      7      |     2     |    2     |          8          |   21    |   15    |   25    |         |         |         |\n",
      "|     33     |      7      |     3     |    2     |          8          |   21    |   15    |   17    |   25    |         |         |\n",
      "|     34     |      7      |     3     |    2     |          8          |   21    |   15    |   25    |   17    |         |         |\n",
      "|     35     |      7      |     2     |    2     |          8          |   21    |   17    |   25    |         |         |         |\n",
      "|     36     |      7      |     2     |    2     |          8          |   21    |   25    |   17    |         |         |         |\n",
      "|     37     |      7      |     2     |    2     |          8          |   15    |   17    |   25    |         |         |         |\n",
      "|     38     |      7      |     2     |    2     |          8          |   15    |   25    |   17    |         |         |         |\n",
      "|     39     |      6      |     2     |    0     |          2          |    3    |   23    |    5    |         |         |         |\n",
      "|     40     |      6      |     2     |    0     |          2          |   23    |    3    |    5    |         |         |         |\n",
      "|     41     |      6      |     2     |    1     |          8          |   24    |    9    |   19    |         |         |         |\n",
      "|     42     |      6      |     2     |    1     |          8          |    7    |    9    |   19    |         |         |         |\n",
      "|     43     |      6      |     3     |    1     |          8          |   24    |    7    |    9    |   19    |         |         |\n",
      "|     44     |      6      |     3     |    1     |          8          |    7    |   24    |    9    |   19    |         |         |\n",
      "|     45     |      6      |     2     |    1     |          8          |   24    |    7    |    9    |         |         |         |\n",
      "|     46     |      6      |     2     |    1     |          8          |    7    |   24    |    9    |         |         |         |\n",
      "|     47     |      6      |     2     |    1     |          8          |   24    |    7    |   19    |         |         |         |\n",
      "|     48     |      6      |     2     |    1     |          8          |    7    |   24    |   19    |         |         |         |\n",
      "|     49     |      6      |     2     |    2     |         16          |   21    |   15    |   17    |         |         |         |\n",
      "|     50     |      6      |     2     |    2     |         16          |   21    |   15    |   16    |         |         |         |\n",
      "|     51     |      6      |     2     |    2     |         16          |   21    |   15    |   18    |         |         |         |\n",
      "|     52     |      6      |     3     |    2     |         16          |   21    |   15    |   17    |   18    |         |         |\n",
      "|     53     |      6      |     3     |    2     |         16          |   21    |   15    |   16    |   17    |         |         |\n",
      "|     54     |      6      |     2     |    2     |         16          |   21    |   17    |   18    |         |         |         |\n",
      "|     55     |      6      |     2     |    2     |         16          |   21    |   16    |   17    |         |         |         |\n",
      "|     56     |      6      |     2     |    2     |         16          |   15    |   17    |   18    |         |         |         |\n",
      "|     57     |      6      |     2     |    2     |         16          |   15    |   16    |   17    |         |         |         |\n",
      "|     58     |      6      |     3     |    2     |         16          |   21    |   15    |   16    |   18    |         |         |\n",
      "|     59     |      6      |     4     |    2     |         16          |   21    |   15    |   16    |   17    |   18    |         |\n",
      "|     60     |      6      |     3     |    2     |         16          |   21    |   16    |   17    |   18    |         |         |\n",
      "|     61     |      6      |     3     |    2     |         16          |   15    |   16    |   17    |   18    |         |         |\n",
      "|     62     |      6      |     2     |    2     |         16          |   21    |   16    |   18    |         |         |         |\n",
      "|     63     |      6      |     2     |    2     |         16          |   15    |   16    |   18    |         |         |         |\n",
      "|     64     |      6      |     2     |    2     |         16          |   16    |   17    |   18    |         |         |         |\n",
      "|     65     |      5      |     2     |    0     |          2          |    3    |   23    |    5    |         |         |         |\n",
      "|     66     |      5      |     2     |    0     |          2          |   23    |    3    |    5    |         |         |         |\n",
      "|     67     |      5      |     2     |    1     |         16          |    8    |    9    |   19    |         |         |         |\n",
      "|     68     |      5      |     2     |    1     |         16          |    6    |    9    |   19    |         |         |         |\n",
      "|     69     |      5      |     2     |    1     |         16          |    7    |    9    |   19    |         |         |         |\n",
      "|     70     |      5      |     3     |    1     |         16          |    6    |    7    |    9    |   19    |         |         |\n",
      "|     71     |      5      |     3     |    1     |         16          |    7    |    8    |    9    |   19    |         |         |\n",
      "|     72     |      5      |     2     |    1     |         16          |    6    |    7    |    9    |         |         |         |\n",
      "|     73     |      5      |     2     |    1     |         16          |    7    |    8    |    9    |         |         |         |\n",
      "|     74     |      5      |     2     |    1     |         16          |    6    |    7    |   19    |         |         |         |\n",
      "|     75     |      5      |     2     |    1     |         16          |    7    |    8    |   19    |         |         |         |\n",
      "|     76     |      5      |     3     |    1     |         16          |    6    |    8    |    9    |   19    |         |         |\n",
      "|     77     |      5      |     4     |    1     |         16          |    6    |    7    |    8    |    9    |   19    |         |\n",
      "|     78     |      5      |     3     |    1     |         16          |    6    |    7    |    8    |    9    |         |         |\n",
      "|     79     |      5      |     3     |    1     |         16          |    6    |    7    |    8    |   19    |         |         |\n",
      "|     80     |      5      |     2     |    1     |         16          |    6    |    8    |    9    |         |         |         |\n",
      "|     81     |      5      |     2     |    1     |         16          |    6    |    8    |   19    |         |         |         |\n",
      "|     82     |      5      |     2     |    1     |         16          |    6    |    7    |    8    |         |         |         |\n",
      "|     83     |      5      |     2     |    2     |         16          |   21    |   15    |   17    |         |         |         |\n",
      "|     84     |      5      |     2     |    2     |         16          |   21    |   15    |   16    |         |         |         |\n",
      "|     85     |      5      |     2     |    2     |         16          |   21    |   15    |   18    |         |         |         |\n",
      "|     86     |      5      |     3     |    2     |         16          |   21    |   15    |   17    |   18    |         |         |\n",
      "|     87     |      5      |     3     |    2     |         16          |   21    |   15    |   16    |   17    |         |         |\n",
      "|     88     |      5      |     2     |    2     |         16          |   21    |   17    |   18    |         |         |         |\n",
      "|     89     |      5      |     2     |    2     |         16          |   21    |   16    |   17    |         |         |         |\n",
      "|     90     |      5      |     2     |    2     |         16          |   15    |   17    |   18    |         |         |         |\n",
      "|     91     |      5      |     2     |    2     |         16          |   15    |   16    |   17    |         |         |         |\n",
      "|     92     |      5      |     3     |    2     |         16          |   21    |   15    |   16    |   18    |         |         |\n",
      "|     93     |      5      |     4     |    2     |         16          |   21    |   15    |   16    |   17    |   18    |         |\n",
      "|     94     |      5      |     3     |    2     |         16          |   21    |   16    |   17    |   18    |         |         |\n",
      "|     95     |      5      |     3     |    2     |         16          |   15    |   16    |   17    |   18    |         |         |\n",
      "|     96     |      5      |     2     |    2     |         16          |   21    |   16    |   18    |         |         |         |\n",
      "|     97     |      5      |     2     |    2     |         16          |   15    |   16    |   18    |         |         |         |\n",
      "|     98     |      5      |     2     |    2     |         16          |   16    |   17    |   18    |         |         |         |\n",
      "|     99     |      4      |     2     |    0     |         11          |    3    |   22    |    5    |         |         |         |\n",
      "|    100     |      4      |     2     |    0     |         11          |   20    |    3    |    5    |         |         |         |\n",
      "|    101     |      4      |     2     |    0     |         11          |   22    |    3    |    5    |         |         |         |\n",
      "|    102     |      4      |     3     |    0     |         11          |   20    |    3    |   22    |    5    |         |         |\n",
      "|    103     |      4      |     3     |    0     |         11          |   20    |   22    |    3    |    5    |         |         |\n",
      "|    104     |      4      |     3     |    0     |         11          |   22    |   20    |    3    |    5    |         |         |\n",
      "|    105     |      4      |     2     |    0     |         11          |   20    |   22    |    3    |         |         |         |\n",
      "|    106     |      4      |     2     |    0     |         11          |   22    |   20    |    3    |         |         |         |\n",
      "|    107     |      4      |     2     |    0     |         11          |   20    |   22    |    5    |         |         |         |\n",
      "|    108     |      4      |     2     |    0     |         11          |   22    |   20    |    5    |         |         |         |\n",
      "|    109     |      4      |     2     |    0     |         11          |   20    |    3    |   22    |         |         |         |\n",
      "|    110     |      4      |     2     |    1     |         16          |    8    |    9    |   19    |         |         |         |\n",
      "|    111     |      4      |     2     |    1     |         16          |    6    |    9    |   19    |         |         |         |\n",
      "|    112     |      4      |     2     |    1     |         16          |    7    |    9    |   19    |         |         |         |\n",
      "|    113     |      4      |     3     |    1     |         16          |    6    |    7    |    9    |   19    |         |         |\n",
      "|    114     |      4      |     3     |    1     |         16          |    7    |    8    |    9    |   19    |         |         |\n",
      "|    115     |      4      |     2     |    1     |         16          |    6    |    7    |    9    |         |         |         |\n",
      "|    116     |      4      |     2     |    1     |         16          |    7    |    8    |    9    |         |         |         |\n",
      "|    117     |      4      |     2     |    1     |         16          |    6    |    7    |   19    |         |         |         |\n",
      "|    118     |      4      |     2     |    1     |         16          |    7    |    8    |   19    |         |         |         |\n",
      "|    119     |      4      |     3     |    1     |         16          |    6    |    8    |    9    |   19    |         |         |\n",
      "|    120     |      4      |     4     |    1     |         16          |    6    |    7    |    8    |    9    |   19    |         |\n",
      "|    121     |      4      |     3     |    1     |         16          |    6    |    7    |    8    |    9    |         |         |\n",
      "|    122     |      4      |     3     |    1     |         16          |    6    |    7    |    8    |   19    |         |         |\n",
      "|    123     |      4      |     2     |    1     |         16          |    6    |    8    |    9    |         |         |         |\n",
      "|    124     |      4      |     2     |    1     |         16          |    6    |    8    |   19    |         |         |         |\n",
      "|    125     |      4      |     2     |    1     |         16          |    6    |    7    |    8    |         |         |         |\n",
      "|    126     |      4      |     2     |    2     |         16          |   21    |   15    |   17    |         |         |         |\n",
      "|    127     |      4      |     2     |    2     |         16          |   21    |   15    |   16    |         |         |         |\n",
      "|    128     |      4      |     2     |    2     |         16          |   21    |   15    |   18    |         |         |         |\n",
      "|    129     |      4      |     3     |    2     |         16          |   21    |   15    |   17    |   18    |         |         |\n",
      "|    130     |      4      |     3     |    2     |         16          |   21    |   15    |   16    |   17    |         |         |\n",
      "|    131     |      4      |     2     |    2     |         16          |   21    |   17    |   18    |         |         |         |\n",
      "|    132     |      4      |     2     |    2     |         16          |   21    |   16    |   17    |         |         |         |\n",
      "|    133     |      4      |     2     |    2     |         16          |   15    |   17    |   18    |         |         |         |\n",
      "|    134     |      4      |     2     |    2     |         16          |   15    |   16    |   17    |         |         |         |\n",
      "|    135     |      4      |     3     |    2     |         16          |   21    |   15    |   16    |   18    |         |         |\n",
      "|    136     |      4      |     4     |    2     |         16          |   21    |   15    |   16    |   17    |   18    |         |\n",
      "|    137     |      4      |     3     |    2     |         16          |   21    |   16    |   17    |   18    |         |         |\n",
      "|    138     |      4      |     3     |    2     |         16          |   15    |   16    |   17    |   18    |         |         |\n",
      "|    139     |      4      |     2     |    2     |         16          |   21    |   16    |   18    |         |         |         |\n",
      "|    140     |      4      |     2     |    2     |         16          |   15    |   16    |   18    |         |         |         |\n",
      "|    141     |      4      |     2     |    2     |         16          |   16    |   17    |   18    |         |         |         |\n",
      "|    142     |      3      |     2     |    0     |         23          |    3    |    4    |    5    |         |         |         |\n",
      "|    143     |      3      |     2     |    0     |         23          |   20    |    3    |    5    |         |         |         |\n",
      "|    144     |      3      |     2     |    0     |         23          |    1    |    3    |    5    |         |         |         |\n",
      "|    145     |      3      |     3     |    0     |         23          |   20    |    3    |    4    |    5    |         |         |\n",
      "|    146     |      3      |     3     |    0     |         23          |   20    |    1    |    3    |    5    |         |         |\n",
      "|    147     |      3      |     3     |    0     |         23          |    1    |   20    |    3    |    5    |         |         |\n",
      "|    148     |      3      |     2     |    0     |         23          |   20    |    1    |    3    |         |         |         |\n",
      "|    149     |      3      |     2     |    0     |         23          |    1    |   20    |    3    |         |         |         |\n",
      "|    150     |      3      |     2     |    0     |         23          |   20    |    1    |    5    |         |         |         |\n",
      "|    151     |      3      |     2     |    0     |         23          |   20    |    4    |    5    |         |         |         |\n",
      "|    152     |      3      |     2     |    0     |         23          |    1    |   20    |    5    |         |         |         |\n",
      "|    153     |      3      |     2     |    0     |         23          |   20    |    3    |    4    |         |         |         |\n",
      "|    154     |      3      |     3     |    0     |         23          |    1    |    3    |    4    |    5    |         |         |\n",
      "|    155     |      3      |     4     |    0     |         23          |    1    |   20    |    3    |    4    |    5    |         |\n",
      "|    156     |      3      |     4     |    0     |         23          |   20    |    1    |    3    |    4    |    5    |         |\n",
      "|    157     |      3      |     3     |    0     |         23          |   20    |    1    |    4    |    5    |         |         |\n",
      "|    158     |      3      |     3     |    0     |         23          |    1    |   20    |    4    |    5    |         |         |\n",
      "|    159     |      3      |     3     |    0     |         23          |    1    |   20    |    3    |    4    |         |         |\n",
      "|    160     |      3      |     3     |    0     |         23          |   20    |    1    |    3    |    4    |         |         |\n",
      "|    161     |      3      |     2     |    0     |         23          |   20    |    1    |    4    |         |         |         |\n",
      "|    162     |      3      |     2     |    0     |         23          |    1    |   20    |    4    |         |         |         |\n",
      "|    163     |      3      |     2     |    0     |         23          |    1    |    4    |    5    |         |         |         |\n",
      "|    164     |      3      |     2     |    0     |         23          |    1    |    3    |    4    |         |         |         |\n",
      "|    165     |      3      |     2     |    1     |         16          |    8    |    9    |   19    |         |         |         |\n",
      "|    166     |      3      |     2     |    1     |         16          |    6    |    9    |   19    |         |         |         |\n",
      "|    167     |      3      |     2     |    1     |         16          |    7    |    9    |   19    |         |         |         |\n",
      "|    168     |      3      |     3     |    1     |         16          |    6    |    7    |    9    |   19    |         |         |\n",
      "|    169     |      3      |     3     |    1     |         16          |    7    |    8    |    9    |   19    |         |         |\n",
      "|    170     |      3      |     2     |    1     |         16          |    6    |    7    |    9    |         |         |         |\n",
      "|    171     |      3      |     2     |    1     |         16          |    7    |    8    |    9    |         |         |         |\n",
      "|    172     |      3      |     2     |    1     |         16          |    6    |    7    |   19    |         |         |         |\n",
      "|    173     |      3      |     2     |    1     |         16          |    7    |    8    |   19    |         |         |         |\n",
      "|    174     |      3      |     3     |    1     |         16          |    6    |    8    |    9    |   19    |         |         |\n",
      "|    175     |      3      |     4     |    1     |         16          |    6    |    7    |    8    |    9    |   19    |         |\n",
      "|    176     |      3      |     3     |    1     |         16          |    6    |    7    |    8    |    9    |         |         |\n",
      "|    177     |      3      |     3     |    1     |         16          |    6    |    7    |    8    |   19    |         |         |\n",
      "|    178     |      3      |     2     |    1     |         16          |    6    |    8    |    9    |         |         |         |\n",
      "|    179     |      3      |     2     |    1     |         16          |    6    |    8    |   19    |         |         |         |\n",
      "|    180     |      3      |     2     |    1     |         16          |    6    |    7    |    8    |         |         |         |\n",
      "|    181     |      3      |     2     |    2     |         16          |   21    |   15    |   17    |         |         |         |\n",
      "|    182     |      3      |     2     |    2     |         16          |   21    |   15    |   16    |         |         |         |\n",
      "|    183     |      3      |     2     |    2     |         16          |   21    |   15    |   18    |         |         |         |\n",
      "|    184     |      3      |     3     |    2     |         16          |   21    |   15    |   17    |   18    |         |         |\n",
      "|    185     |      3      |     3     |    2     |         16          |   21    |   15    |   16    |   17    |         |         |\n",
      "|    186     |      3      |     2     |    2     |         16          |   21    |   17    |   18    |         |         |         |\n",
      "|    187     |      3      |     2     |    2     |         16          |   21    |   16    |   17    |         |         |         |\n",
      "|    188     |      3      |     2     |    2     |         16          |   15    |   17    |   18    |         |         |         |\n",
      "|    189     |      3      |     2     |    2     |         16          |   15    |   16    |   17    |         |         |         |\n",
      "|    190     |      3      |     3     |    2     |         16          |   21    |   15    |   16    |   18    |         |         |\n",
      "|    191     |      3      |     4     |    2     |         16          |   21    |   15    |   16    |   17    |   18    |         |\n",
      "|    192     |      3      |     3     |    2     |         16          |   21    |   16    |   17    |   18    |         |         |\n",
      "|    193     |      3      |     3     |    2     |         16          |   15    |   16    |   17    |   18    |         |         |\n",
      "|    194     |      3      |     2     |    2     |         16          |   21    |   16    |   18    |         |         |         |\n",
      "|    195     |      3      |     2     |    2     |         16          |   15    |   16    |   18    |         |         |         |\n",
      "|    196     |      3      |     2     |    2     |         16          |   16    |   17    |   18    |         |         |         |\n",
      "|    197     |      2      |     2     |    0     |         23          |    3    |    4    |    5    |         |         |         |\n",
      "|    198     |      2      |     2     |    0     |         23          |   20    |    3    |    5    |         |         |         |\n",
      "|    199     |      2      |     2     |    0     |         23          |    1    |    3    |    5    |         |         |         |\n",
      "|    200     |      2      |     3     |    0     |         23          |   20    |    3    |    4    |    5    |         |         |\n",
      "|    201     |      2      |     3     |    0     |         23          |   20    |    1    |    3    |    5    |         |         |\n",
      "|    202     |      2      |     3     |    0     |         23          |    1    |   20    |    3    |    5    |         |         |\n",
      "|    203     |      2      |     2     |    0     |         23          |   20    |    1    |    3    |         |         |         |\n",
      "|    204     |      2      |     2     |    0     |         23          |    1    |   20    |    3    |         |         |         |\n",
      "|    205     |      2      |     2     |    0     |         23          |   20    |    1    |    5    |         |         |         |\n",
      "|    206     |      2      |     2     |    0     |         23          |   20    |    4    |    5    |         |         |         |\n",
      "|    207     |      2      |     2     |    0     |         23          |    1    |   20    |    5    |         |         |         |\n",
      "|    208     |      2      |     2     |    0     |         23          |   20    |    3    |    4    |         |         |         |\n",
      "|    209     |      2      |     3     |    0     |         23          |    1    |    3    |    4    |    5    |         |         |\n",
      "|    210     |      2      |     4     |    0     |         23          |    1    |   20    |    3    |    4    |    5    |         |\n",
      "|    211     |      2      |     4     |    0     |         23          |   20    |    1    |    3    |    4    |    5    |         |\n",
      "|    212     |      2      |     3     |    0     |         23          |   20    |    1    |    4    |    5    |         |         |\n",
      "|    213     |      2      |     3     |    0     |         23          |    1    |   20    |    4    |    5    |         |         |\n",
      "|    214     |      2      |     3     |    0     |         23          |    1    |   20    |    3    |    4    |         |         |\n",
      "|    215     |      2      |     3     |    0     |         23          |   20    |    1    |    3    |    4    |         |         |\n",
      "|    216     |      2      |     2     |    0     |         23          |   20    |    1    |    4    |         |         |         |\n",
      "|    217     |      2      |     2     |    0     |         23          |    1    |   20    |    4    |         |         |         |\n",
      "|    218     |      2      |     2     |    0     |         23          |    1    |    4    |    5    |         |         |         |\n",
      "|    219     |      2      |     2     |    0     |         23          |    1    |    3    |    4    |         |         |         |\n",
      "|    220     |      2      |     2     |    1     |         16          |    8    |    9    |   19    |         |         |         |\n",
      "|    221     |      2      |     2     |    1     |         16          |    6    |    9    |   19    |         |         |         |\n",
      "|    222     |      2      |     2     |    1     |         16          |    7    |    9    |   19    |         |         |         |\n",
      "|    223     |      2      |     3     |    1     |         16          |    6    |    7    |    9    |   19    |         |         |\n",
      "|    224     |      2      |     3     |    1     |         16          |    7    |    8    |    9    |   19    |         |         |\n",
      "|    225     |      2      |     2     |    1     |         16          |    6    |    7    |    9    |         |         |         |\n",
      "|    226     |      2      |     2     |    1     |         16          |    7    |    8    |    9    |         |         |         |\n",
      "|    227     |      2      |     2     |    1     |         16          |    6    |    7    |   19    |         |         |         |\n",
      "|    228     |      2      |     2     |    1     |         16          |    7    |    8    |   19    |         |         |         |\n",
      "|    229     |      2      |     3     |    1     |         16          |    6    |    8    |    9    |   19    |         |         |\n",
      "|    230     |      2      |     4     |    1     |         16          |    6    |    7    |    8    |    9    |   19    |         |\n",
      "|    231     |      2      |     3     |    1     |         16          |    6    |    7    |    8    |    9    |         |         |\n",
      "|    232     |      2      |     3     |    1     |         16          |    6    |    7    |    8    |   19    |         |         |\n",
      "|    233     |      2      |     2     |    1     |         16          |    6    |    8    |    9    |         |         |         |\n",
      "|    234     |      2      |     2     |    1     |         16          |    6    |    8    |   19    |         |         |         |\n",
      "|    235     |      2      |     2     |    1     |         16          |    6    |    7    |    8    |         |         |         |\n",
      "|    236     |      2      |     2     |    2     |         42          |   13    |   15    |   17    |         |         |         |\n",
      "|    237     |      2      |     2     |    2     |         42          |   14    |   15    |   17    |         |         |         |\n",
      "|    238     |      2      |     2     |    2     |         42          |   13    |   15    |   16    |         |         |         |\n",
      "|    239     |      2      |     2     |    2     |         42          |   14    |   15    |   16    |         |         |         |\n",
      "|    240     |      2      |     2     |    2     |         42          |   13    |   15    |   18    |         |         |         |\n",
      "|    241     |      2      |     2     |    2     |         42          |   14    |   15    |   18    |         |         |         |\n",
      "|    242     |      2      |     3     |    2     |         42          |   13    |   15    |   17    |   18    |         |         |\n",
      "|    243     |      2      |     3     |    2     |         42          |   14    |   15    |   17    |   18    |         |         |\n",
      "|    244     |      2      |     3     |    2     |         42          |   13    |   15    |   16    |   17    |         |         |\n",
      "|    245     |      2      |     3     |    2     |         42          |   14    |   15    |   16    |   17    |         |         |\n",
      "|    246     |      2      |     2     |    2     |         42          |   13    |   17    |   18    |         |         |         |\n",
      "|    247     |      2      |     2     |    2     |         42          |   14    |   17    |   18    |         |         |         |\n",
      "|    248     |      2      |     2     |    2     |         42          |   13    |   16    |   17    |         |         |         |\n",
      "|    249     |      2      |     2     |    2     |         42          |   14    |   16    |   17    |         |         |         |\n",
      "|    250     |      2      |     2     |    2     |         42          |   15    |   17    |   18    |         |         |         |\n",
      "|    251     |      2      |     2     |    2     |         42          |   15    |   16    |   17    |         |         |         |\n",
      "|    252     |      2      |     3     |    2     |         42          |   13    |   15    |   16    |   18    |         |         |\n",
      "|    253     |      2      |     3     |    2     |         42          |   14    |   15    |   16    |   18    |         |         |\n",
      "|    254     |      2      |     4     |    2     |         42          |   13    |   15    |   16    |   17    |   18    |         |\n",
      "|    255     |      2      |     4     |    2     |         42          |   14    |   15    |   16    |   17    |   18    |         |\n",
      "|    256     |      2      |     3     |    2     |         42          |   13    |   16    |   17    |   18    |         |         |\n",
      "|    257     |      2      |     3     |    2     |         42          |   14    |   16    |   17    |   18    |         |         |\n",
      "|    258     |      2      |     3     |    2     |         42          |   15    |   16    |   17    |   18    |         |         |\n",
      "|    259     |      2      |     2     |    2     |         42          |   13    |   16    |   18    |         |         |         |\n",
      "|    260     |      2      |     2     |    2     |         42          |   14    |   16    |   18    |         |         |         |\n",
      "|    261     |      2      |     2     |    2     |         42          |   15    |   16    |   18    |         |         |         |\n",
      "|    262     |      2      |     2     |    2     |         42          |   16    |   17    |   18    |         |         |         |\n",
      "|    263     |      2      |     3     |    2     |         42          |   13    |   14    |   15    |   17    |         |         |\n",
      "|    264     |      2      |     3     |    2     |         42          |   13    |   14    |   15    |   16    |         |         |\n",
      "|    265     |      2      |     3     |    2     |         42          |   13    |   14    |   15    |   18    |         |         |\n",
      "|    266     |      2      |     4     |    2     |         42          |   13    |   14    |   15    |   17    |   18    |         |\n",
      "|    267     |      2      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |         |\n",
      "|    268     |      2      |     3     |    2     |         42          |   13    |   14    |   17    |   18    |         |         |\n",
      "|    269     |      2      |     3     |    2     |         42          |   13    |   14    |   16    |   17    |         |         |\n",
      "|    270     |      2      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   18    |         |\n",
      "|    271     |      2      |     5     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |   18    |\n",
      "|    272     |      2      |     4     |    2     |         42          |   13    |   14    |   16    |   17    |   18    |         |\n",
      "|    273     |      2      |     3     |    2     |         42          |   13    |   14    |   16    |   18    |         |         |\n",
      "|    274     |      2      |     2     |    2     |         42          |   13    |   14    |   15    |         |         |         |\n",
      "|    275     |      2      |     2     |    2     |         42          |   13    |   14    |   16    |         |         |         |\n",
      "|    276     |      2      |     2     |    2     |         42          |   13    |   14    |   17    |         |         |         |\n",
      "|    277     |      2      |     2     |    2     |         42          |   13    |   14    |   18    |         |         |         |\n",
      "|    278     |      1      |     2     |    0     |         42          |    3    |    4    |    5    |         |         |         |\n",
      "|    279     |      1      |     2     |    0     |         42          |    0    |    3    |    5    |         |         |         |\n",
      "|    280     |      1      |     2     |    0     |         42          |    2    |    3    |    5    |         |         |         |\n",
      "|    281     |      1      |     2     |    0     |         42          |    1    |    3    |    5    |         |         |         |\n",
      "|    282     |      1      |     3     |    0     |         42          |    0    |    3    |    4    |    5    |         |         |\n",
      "|    283     |      1      |     3     |    0     |         42          |    2    |    3    |    4    |    5    |         |         |\n",
      "|    284     |      1      |     3     |    0     |         42          |    0    |    1    |    3    |    5    |         |         |\n",
      "|    285     |      1      |     3     |    0     |         42          |    1    |    2    |    3    |    5    |         |         |\n",
      "|    286     |      1      |     2     |    0     |         42          |    0    |    1    |    3    |         |         |         |\n",
      "|    287     |      1      |     2     |    0     |         42          |    1    |    2    |    3    |         |         |         |\n",
      "|    288     |      1      |     2     |    0     |         42          |    0    |    1    |    5    |         |         |         |\n",
      "|    289     |      1      |     2     |    0     |         42          |    0    |    4    |    5    |         |         |         |\n",
      "|    290     |      1      |     2     |    0     |         42          |    2    |    4    |    5    |         |         |         |\n",
      "|    291     |      1      |     2     |    0     |         42          |    1    |    2    |    5    |         |         |         |\n",
      "|    292     |      1      |     2     |    0     |         42          |    0    |    3    |    4    |         |         |         |\n",
      "|    293     |      1      |     2     |    0     |         42          |    2    |    3    |    4    |         |         |         |\n",
      "|    294     |      1      |     3     |    0     |         42          |    1    |    3    |    4    |    5    |         |         |\n",
      "|    295     |      1      |     4     |    0     |         42          |    1    |    2    |    3    |    4    |    5    |         |\n",
      "|    296     |      1      |     4     |    0     |         42          |    0    |    1    |    3    |    4    |    5    |         |\n",
      "|    297     |      1      |     3     |    0     |         42          |    0    |    1    |    4    |    5    |         |         |\n",
      "|    298     |      1      |     3     |    0     |         42          |    1    |    2    |    4    |    5    |         |         |\n",
      "|    299     |      1      |     3     |    0     |         42          |    1    |    2    |    3    |    4    |         |         |\n",
      "|    300     |      1      |     3     |    0     |         42          |    0    |    1    |    3    |    4    |         |         |\n",
      "|    301     |      1      |     2     |    0     |         42          |    0    |    1    |    4    |         |         |         |\n",
      "|    302     |      1      |     2     |    0     |         42          |    1    |    2    |    4    |         |         |         |\n",
      "|    303     |      1      |     2     |    0     |         42          |    1    |    4    |    5    |         |         |         |\n",
      "|    304     |      1      |     2     |    0     |         42          |    1    |    3    |    4    |         |         |         |\n",
      "|    305     |      1      |     3     |    0     |         42          |    0    |    2    |    3    |    5    |         |         |\n",
      "|    306     |      1      |     4     |    0     |         42          |    0    |    2    |    3    |    4    |    5    |         |\n",
      "|    307     |      1      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    5    |         |\n",
      "|    308     |      1      |     3     |    0     |         42          |    0    |    1    |    2    |    3    |         |         |\n",
      "|    309     |      1      |     3     |    0     |         42          |    0    |    1    |    2    |    5    |         |         |\n",
      "|    310     |      1      |     3     |    0     |         42          |    0    |    2    |    4    |    5    |         |         |\n",
      "|    311     |      1      |     3     |    0     |         42          |    0    |    2    |    3    |    4    |         |         |\n",
      "|    312     |      1      |     5     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |    5    |\n",
      "|    313     |      1      |     4     |    0     |         42          |    0    |    1    |    2    |    4    |    5    |         |\n",
      "|    314     |      1      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |         |\n",
      "|    315     |      1      |     3     |    0     |         42          |    0    |    1    |    2    |    4    |         |         |\n",
      "|    316     |      1      |     2     |    0     |         42          |    0    |    2    |    3    |         |         |         |\n",
      "|    317     |      1      |     2     |    0     |         42          |    0    |    2    |    4    |         |         |         |\n",
      "|    318     |      1      |     2     |    0     |         42          |    0    |    2    |    5    |         |         |         |\n",
      "|    319     |      1      |     2     |    0     |         42          |    0    |    1    |    2    |         |         |         |\n",
      "|    320     |      1      |     2     |    1     |         16          |    8    |    9    |   19    |         |         |         |\n",
      "|    321     |      1      |     2     |    1     |         16          |    6    |    9    |   19    |         |         |         |\n",
      "|    322     |      1      |     2     |    1     |         16          |    7    |    9    |   19    |         |         |         |\n",
      "|    323     |      1      |     3     |    1     |         16          |    6    |    7    |    9    |   19    |         |         |\n",
      "|    324     |      1      |     3     |    1     |         16          |    7    |    8    |    9    |   19    |         |         |\n",
      "|    325     |      1      |     2     |    1     |         16          |    6    |    7    |    9    |         |         |         |\n",
      "|    326     |      1      |     2     |    1     |         16          |    7    |    8    |    9    |         |         |         |\n",
      "|    327     |      1      |     2     |    1     |         16          |    6    |    7    |   19    |         |         |         |\n",
      "|    328     |      1      |     2     |    1     |         16          |    7    |    8    |   19    |         |         |         |\n",
      "|    329     |      1      |     3     |    1     |         16          |    6    |    8    |    9    |   19    |         |         |\n",
      "|    330     |      1      |     4     |    1     |         16          |    6    |    7    |    8    |    9    |   19    |         |\n",
      "|    331     |      1      |     3     |    1     |         16          |    6    |    7    |    8    |    9    |         |         |\n",
      "|    332     |      1      |     3     |    1     |         16          |    6    |    7    |    8    |   19    |         |         |\n",
      "|    333     |      1      |     2     |    1     |         16          |    6    |    8    |    9    |         |         |         |\n",
      "|    334     |      1      |     2     |    1     |         16          |    6    |    8    |   19    |         |         |         |\n",
      "|    335     |      1      |     2     |    1     |         16          |    6    |    7    |    8    |         |         |         |\n",
      "|    336     |      1      |     2     |    2     |         42          |   13    |   15    |   17    |         |         |         |\n",
      "|    337     |      1      |     2     |    2     |         42          |   14    |   15    |   17    |         |         |         |\n",
      "|    338     |      1      |     2     |    2     |         42          |   13    |   15    |   16    |         |         |         |\n",
      "|    339     |      1      |     2     |    2     |         42          |   14    |   15    |   16    |         |         |         |\n",
      "|    340     |      1      |     2     |    2     |         42          |   13    |   15    |   18    |         |         |         |\n",
      "|    341     |      1      |     2     |    2     |         42          |   14    |   15    |   18    |         |         |         |\n",
      "|    342     |      1      |     3     |    2     |         42          |   13    |   15    |   17    |   18    |         |         |\n",
      "|    343     |      1      |     3     |    2     |         42          |   14    |   15    |   17    |   18    |         |         |\n",
      "|    344     |      1      |     3     |    2     |         42          |   13    |   15    |   16    |   17    |         |         |\n",
      "|    345     |      1      |     3     |    2     |         42          |   14    |   15    |   16    |   17    |         |         |\n",
      "|    346     |      1      |     2     |    2     |         42          |   13    |   17    |   18    |         |         |         |\n",
      "|    347     |      1      |     2     |    2     |         42          |   14    |   17    |   18    |         |         |         |\n",
      "|    348     |      1      |     2     |    2     |         42          |   13    |   16    |   17    |         |         |         |\n",
      "|    349     |      1      |     2     |    2     |         42          |   14    |   16    |   17    |         |         |         |\n",
      "|    350     |      1      |     2     |    2     |         42          |   15    |   17    |   18    |         |         |         |\n",
      "|    351     |      1      |     2     |    2     |         42          |   15    |   16    |   17    |         |         |         |\n",
      "|    352     |      1      |     3     |    2     |         42          |   13    |   15    |   16    |   18    |         |         |\n",
      "|    353     |      1      |     3     |    2     |         42          |   14    |   15    |   16    |   18    |         |         |\n",
      "|    354     |      1      |     4     |    2     |         42          |   13    |   15    |   16    |   17    |   18    |         |\n",
      "|    355     |      1      |     4     |    2     |         42          |   14    |   15    |   16    |   17    |   18    |         |\n",
      "|    356     |      1      |     3     |    2     |         42          |   13    |   16    |   17    |   18    |         |         |\n",
      "|    357     |      1      |     3     |    2     |         42          |   14    |   16    |   17    |   18    |         |         |\n",
      "|    358     |      1      |     3     |    2     |         42          |   15    |   16    |   17    |   18    |         |         |\n",
      "|    359     |      1      |     2     |    2     |         42          |   13    |   16    |   18    |         |         |         |\n",
      "|    360     |      1      |     2     |    2     |         42          |   14    |   16    |   18    |         |         |         |\n",
      "|    361     |      1      |     2     |    2     |         42          |   15    |   16    |   18    |         |         |         |\n",
      "|    362     |      1      |     2     |    2     |         42          |   16    |   17    |   18    |         |         |         |\n",
      "|    363     |      1      |     3     |    2     |         42          |   13    |   14    |   15    |   17    |         |         |\n",
      "|    364     |      1      |     3     |    2     |         42          |   13    |   14    |   15    |   16    |         |         |\n",
      "|    365     |      1      |     3     |    2     |         42          |   13    |   14    |   15    |   18    |         |         |\n",
      "|    366     |      1      |     4     |    2     |         42          |   13    |   14    |   15    |   17    |   18    |         |\n",
      "|    367     |      1      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |         |\n",
      "|    368     |      1      |     3     |    2     |         42          |   13    |   14    |   17    |   18    |         |         |\n",
      "|    369     |      1      |     3     |    2     |         42          |   13    |   14    |   16    |   17    |         |         |\n",
      "|    370     |      1      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   18    |         |\n",
      "|    371     |      1      |     5     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |   18    |\n",
      "|    372     |      1      |     4     |    2     |         42          |   13    |   14    |   16    |   17    |   18    |         |\n",
      "|    373     |      1      |     3     |    2     |         42          |   13    |   14    |   16    |   18    |         |         |\n",
      "|    374     |      1      |     2     |    2     |         42          |   13    |   14    |   15    |         |         |         |\n",
      "|    375     |      1      |     2     |    2     |         42          |   13    |   14    |   16    |         |         |         |\n",
      "|    376     |      1      |     2     |    2     |         42          |   13    |   14    |   17    |         |         |         |\n",
      "|    377     |      1      |     2     |    2     |         42          |   13    |   14    |   18    |         |         |         |\n",
      "|    378     |      0      |     2     |    0     |         42          |    3    |    4    |    5    |         |         |         |\n",
      "|    379     |      0      |     2     |    0     |         42          |    0    |    3    |    5    |         |         |         |\n",
      "|    380     |      0      |     2     |    0     |         42          |    2    |    3    |    5    |         |         |         |\n",
      "|    381     |      0      |     2     |    0     |         42          |    1    |    3    |    5    |         |         |         |\n",
      "|    382     |      0      |     3     |    0     |         42          |    0    |    3    |    4    |    5    |         |         |\n",
      "|    383     |      0      |     3     |    0     |         42          |    2    |    3    |    4    |    5    |         |         |\n",
      "|    384     |      0      |     3     |    0     |         42          |    0    |    1    |    3    |    5    |         |         |\n",
      "|    385     |      0      |     3     |    0     |         42          |    1    |    2    |    3    |    5    |         |         |\n",
      "|    386     |      0      |     2     |    0     |         42          |    0    |    1    |    3    |         |         |         |\n",
      "|    387     |      0      |     2     |    0     |         42          |    1    |    2    |    3    |         |         |         |\n",
      "|    388     |      0      |     2     |    0     |         42          |    0    |    1    |    5    |         |         |         |\n",
      "|    389     |      0      |     2     |    0     |         42          |    0    |    4    |    5    |         |         |         |\n",
      "|    390     |      0      |     2     |    0     |         42          |    2    |    4    |    5    |         |         |         |\n",
      "|    391     |      0      |     2     |    0     |         42          |    1    |    2    |    5    |         |         |         |\n",
      "|    392     |      0      |     2     |    0     |         42          |    0    |    3    |    4    |         |         |         |\n",
      "|    393     |      0      |     2     |    0     |         42          |    2    |    3    |    4    |         |         |         |\n",
      "|    394     |      0      |     3     |    0     |         42          |    1    |    3    |    4    |    5    |         |         |\n",
      "|    395     |      0      |     4     |    0     |         42          |    1    |    2    |    3    |    4    |    5    |         |\n",
      "|    396     |      0      |     4     |    0     |         42          |    0    |    1    |    3    |    4    |    5    |         |\n",
      "|    397     |      0      |     3     |    0     |         42          |    0    |    1    |    4    |    5    |         |         |\n",
      "|    398     |      0      |     3     |    0     |         42          |    1    |    2    |    4    |    5    |         |         |\n",
      "|    399     |      0      |     3     |    0     |         42          |    1    |    2    |    3    |    4    |         |         |\n",
      "|    400     |      0      |     3     |    0     |         42          |    0    |    1    |    3    |    4    |         |         |\n",
      "|    401     |      0      |     2     |    0     |         42          |    0    |    1    |    4    |         |         |         |\n",
      "|    402     |      0      |     2     |    0     |         42          |    1    |    2    |    4    |         |         |         |\n",
      "|    403     |      0      |     2     |    0     |         42          |    1    |    4    |    5    |         |         |         |\n",
      "|    404     |      0      |     2     |    0     |         42          |    1    |    3    |    4    |         |         |         |\n",
      "|    405     |      0      |     3     |    0     |         42          |    0    |    2    |    3    |    5    |         |         |\n",
      "|    406     |      0      |     4     |    0     |         42          |    0    |    2    |    3    |    4    |    5    |         |\n",
      "|    407     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    5    |         |\n",
      "|    408     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    3    |         |         |\n",
      "|    409     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    5    |         |         |\n",
      "|    410     |      0      |     3     |    0     |         42          |    0    |    2    |    4    |    5    |         |         |\n",
      "|    411     |      0      |     3     |    0     |         42          |    0    |    2    |    3    |    4    |         |         |\n",
      "|    412     |      0      |     5     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |    5    |\n",
      "|    413     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    4    |    5    |         |\n",
      "|    414     |      0      |     4     |    0     |         42          |    0    |    1    |    2    |    3    |    4    |         |\n",
      "|    415     |      0      |     3     |    0     |         42          |    0    |    1    |    2    |    4    |         |         |\n",
      "|    416     |      0      |     2     |    0     |         42          |    0    |    2    |    3    |         |         |         |\n",
      "|    417     |      0      |     2     |    0     |         42          |    0    |    2    |    4    |         |         |         |\n",
      "|    418     |      0      |     2     |    0     |         42          |    0    |    2    |    5    |         |         |         |\n",
      "|    419     |      0      |     2     |    0     |         42          |    0    |    1    |    2    |         |         |         |\n",
      "|    420     |      0      |     2     |    1     |         42          |    8    |    9    |   10    |         |         |         |\n",
      "|    421     |      0      |     2     |    1     |         42          |    8    |    9    |   11    |         |         |         |\n",
      "|    422     |      0      |     2     |    1     |         42          |    6    |    9    |   10    |         |         |         |\n",
      "|    423     |      0      |     2     |    1     |         42          |    6    |    9    |   11    |         |         |         |\n",
      "|    424     |      0      |     2     |    1     |         42          |    7    |    9    |   10    |         |         |         |\n",
      "|    425     |      0      |     2     |    1     |         42          |    7    |    9    |   11    |         |         |         |\n",
      "|    426     |      0      |     3     |    1     |         42          |    6    |    7    |    9    |   10    |         |         |\n",
      "|    427     |      0      |     3     |    1     |         42          |    6    |    7    |    9    |   11    |         |         |\n",
      "|    428     |      0      |     3     |    1     |         42          |    7    |    8    |    9    |   10    |         |         |\n",
      "|    429     |      0      |     3     |    1     |         42          |    7    |    8    |    9    |   11    |         |         |\n",
      "|    430     |      0      |     2     |    1     |         42          |    6    |    7    |    9    |         |         |         |\n",
      "|    431     |      0      |     2     |    1     |         42          |    7    |    8    |    9    |         |         |         |\n",
      "|    432     |      0      |     2     |    1     |         42          |    6    |    7    |   10    |         |         |         |\n",
      "|    433     |      0      |     2     |    1     |         42          |    6    |    7    |   11    |         |         |         |\n",
      "|    434     |      0      |     2     |    1     |         42          |    7    |    8    |   10    |         |         |         |\n",
      "|    435     |      0      |     2     |    1     |         42          |    7    |    8    |   11    |         |         |         |\n",
      "|    436     |      0      |     3     |    1     |         42          |    6    |    8    |    9    |   10    |         |         |\n",
      "|    437     |      0      |     3     |    1     |         42          |    6    |    8    |    9    |   11    |         |         |\n",
      "|    438     |      0      |     4     |    1     |         42          |    6    |    7    |    8    |    9    |   10    |         |\n",
      "|    439     |      0      |     4     |    1     |         42          |    6    |    7    |    8    |    9    |   11    |         |\n",
      "|    440     |      0      |     3     |    1     |         42          |    6    |    7    |    8    |    9    |         |         |\n",
      "|    441     |      0      |     3     |    1     |         42          |    6    |    7    |    8    |   10    |         |         |\n",
      "|    442     |      0      |     3     |    1     |         42          |    6    |    7    |    8    |   11    |         |         |\n",
      "|    443     |      0      |     2     |    1     |         42          |    6    |    8    |    9    |         |         |         |\n",
      "|    444     |      0      |     2     |    1     |         42          |    6    |    8    |   10    |         |         |         |\n",
      "|    445     |      0      |     2     |    1     |         42          |    6    |    8    |   11    |         |         |         |\n",
      "|    446     |      0      |     2     |    1     |         42          |    6    |    7    |    8    |         |         |         |\n",
      "|    447     |      0      |     3     |    1     |         42          |    8    |    9    |   10    |   11    |         |         |\n",
      "|    448     |      0      |     3     |    1     |         42          |    6    |    9    |   10    |   11    |         |         |\n",
      "|    449     |      0      |     3     |    1     |         42          |    7    |    9    |   10    |   11    |         |         |\n",
      "|    450     |      0      |     4     |    1     |         42          |    6    |    7    |    9    |   10    |   11    |         |\n",
      "|    451     |      0      |     4     |    1     |         42          |    7    |    8    |    9    |   10    |   11    |         |\n",
      "|    452     |      0      |     3     |    1     |         42          |    6    |    7    |   10    |   11    |         |         |\n",
      "|    453     |      0      |     3     |    1     |         42          |    7    |    8    |   10    |   11    |         |         |\n",
      "|    454     |      0      |     4     |    1     |         42          |    6    |    8    |    9    |   10    |   11    |         |\n",
      "|    455     |      0      |     5     |    1     |         42          |    6    |    7    |    8    |    9    |   10    |   11    |\n",
      "|    456     |      0      |     4     |    1     |         42          |    6    |    7    |    8    |   10    |   11    |         |\n",
      "|    457     |      0      |     3     |    1     |         42          |    6    |    8    |   10    |   11    |         |         |\n",
      "|    458     |      0      |     2     |    1     |         42          |    6    |   10    |   11    |         |         |         |\n",
      "|    459     |      0      |     2     |    1     |         42          |    7    |   10    |   11    |         |         |         |\n",
      "|    460     |      0      |     2     |    1     |         42          |    8    |   10    |   11    |         |         |         |\n",
      "|    461     |      0      |     2     |    1     |         42          |    9    |   10    |   11    |         |         |         |\n",
      "|    462     |      0      |     2     |    2     |         42          |   13    |   15    |   17    |         |         |         |\n",
      "|    463     |      0      |     2     |    2     |         42          |   14    |   15    |   17    |         |         |         |\n",
      "|    464     |      0      |     2     |    2     |         42          |   13    |   15    |   16    |         |         |         |\n",
      "|    465     |      0      |     2     |    2     |         42          |   14    |   15    |   16    |         |         |         |\n",
      "|    466     |      0      |     2     |    2     |         42          |   13    |   15    |   18    |         |         |         |\n",
      "|    467     |      0      |     2     |    2     |         42          |   14    |   15    |   18    |         |         |         |\n",
      "|    468     |      0      |     3     |    2     |         42          |   13    |   15    |   17    |   18    |         |         |\n",
      "|    469     |      0      |     3     |    2     |         42          |   14    |   15    |   17    |   18    |         |         |\n",
      "|    470     |      0      |     3     |    2     |         42          |   13    |   15    |   16    |   17    |         |         |\n",
      "|    471     |      0      |     3     |    2     |         42          |   14    |   15    |   16    |   17    |         |         |\n",
      "|    472     |      0      |     2     |    2     |         42          |   13    |   17    |   18    |         |         |         |\n",
      "|    473     |      0      |     2     |    2     |         42          |   14    |   17    |   18    |         |         |         |\n",
      "|    474     |      0      |     2     |    2     |         42          |   13    |   16    |   17    |         |         |         |\n",
      "|    475     |      0      |     2     |    2     |         42          |   14    |   16    |   17    |         |         |         |\n",
      "|    476     |      0      |     2     |    2     |         42          |   15    |   17    |   18    |         |         |         |\n",
      "|    477     |      0      |     2     |    2     |         42          |   15    |   16    |   17    |         |         |         |\n",
      "|    478     |      0      |     3     |    2     |         42          |   13    |   15    |   16    |   18    |         |         |\n",
      "|    479     |      0      |     3     |    2     |         42          |   14    |   15    |   16    |   18    |         |         |\n",
      "|    480     |      0      |     4     |    2     |         42          |   13    |   15    |   16    |   17    |   18    |         |\n",
      "|    481     |      0      |     4     |    2     |         42          |   14    |   15    |   16    |   17    |   18    |         |\n",
      "|    482     |      0      |     3     |    2     |         42          |   13    |   16    |   17    |   18    |         |         |\n",
      "|    483     |      0      |     3     |    2     |         42          |   14    |   16    |   17    |   18    |         |         |\n",
      "|    484     |      0      |     3     |    2     |         42          |   15    |   16    |   17    |   18    |         |         |\n",
      "|    485     |      0      |     2     |    2     |         42          |   13    |   16    |   18    |         |         |         |\n",
      "|    486     |      0      |     2     |    2     |         42          |   14    |   16    |   18    |         |         |         |\n",
      "|    487     |      0      |     2     |    2     |         42          |   15    |   16    |   18    |         |         |         |\n",
      "|    488     |      0      |     2     |    2     |         42          |   16    |   17    |   18    |         |         |         |\n",
      "|    489     |      0      |     3     |    2     |         42          |   13    |   14    |   15    |   17    |         |         |\n",
      "|    490     |      0      |     3     |    2     |         42          |   13    |   14    |   15    |   16    |         |         |\n",
      "|    491     |      0      |     3     |    2     |         42          |   13    |   14    |   15    |   18    |         |         |\n",
      "|    492     |      0      |     4     |    2     |         42          |   13    |   14    |   15    |   17    |   18    |         |\n",
      "|    493     |      0      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |         |\n",
      "|    494     |      0      |     3     |    2     |         42          |   13    |   14    |   17    |   18    |         |         |\n",
      "|    495     |      0      |     3     |    2     |         42          |   13    |   14    |   16    |   17    |         |         |\n",
      "|    496     |      0      |     4     |    2     |         42          |   13    |   14    |   15    |   16    |   18    |         |\n",
      "|    497     |      0      |     5     |    2     |         42          |   13    |   14    |   15    |   16    |   17    |   18    |\n",
      "|    498     |      0      |     4     |    2     |         42          |   13    |   14    |   16    |   17    |   18    |         |\n",
      "|    499     |      0      |     3     |    2     |         42          |   13    |   14    |   16    |   18    |         |         |\n",
      "|    500     |      0      |     2     |    2     |         42          |   13    |   14    |   15    |         |         |         |\n",
      "|    501     |      0      |     2     |    2     |         42          |   13    |   14    |   16    |         |         |         |\n",
      "|    502     |      0      |     2     |    2     |         42          |   13    |   14    |   17    |         |         |         |\n",
      "|    503     |      0      |     2     |    2     |         42          |   13    |   14    |   18    |         |         |         |\n",
      "+------------+-------------+-----------+----------+---------------------+---------+---------+---------+---------+---------+---------+\n"
     ]
    }
   ],
   "source": [
    "mulitpl5_preprocessing._view_db('graph')\n",
    "mulitpl5_preprocessing._view_db('node_classes')\n",
    "mulitpl5_preprocessing._view_db('edge_details')\n",
    "mulitpl5_preprocessing._view_db('simplices')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "id": "5544d03e",
   "metadata": {},
   "outputs": [],
   "source": [
    "class AdjGraphOriginal():\n",
    "    \n",
    "    empty_graph = dgl.heterograph({('node', 'to', 'node'): ([], [])})\n",
    "    \n",
    "    \n",
    "    def __init__(self,\n",
    "        graph = empty_graph):\n",
    "        \n",
    "        assert isinstance(graph, dgl.DGLHeteroGraph), \\\n",
    "        'Keyword argument \\\"graph\\\" of AdjGraph\\'s init methodmust be a dgl.DGLHeteroGraph.'\n",
    "        \n",
    "        self.seed_graph = graph\n",
    "        \n",
    "        #self.seed_nodes = self.seed_graph.nodes()\n",
    "        \n",
    "        src, dst = self.seed_graph.edges()\n",
    "        src = src.tolist()\n",
    "        dst = dst.tolist()\n",
    "        self.seed_edges = (src, dst)\n",
    "                \n",
    "        seed_edge_pairs = []\n",
    "        for i, u in enumerate(src):\n",
    "            v = dst[i]\n",
    "            seed_edge_pairs.append((u,v))\n",
    "        self.seed_edge_pairs = seed_edge_pairs\n",
    "        \n",
    "        set_of_nodes = set(src+dst)\n",
    "        self.seed_nodes = list(set_of_nodes)\n",
    "        \n",
    "        assert ((len(self.seed_edge_pairs) - len(set(self.seed_edge_pairs))) == 0),\\\n",
    "        'Keyword argument \\\"graph\\\" of AdjGraph\\'s init method must not have multiple edges'\n",
    "        \n",
    "        max_vertex_index = max(self.seed_nodes)\n",
    "        self.seed_edge_indices = [max_vertex_index+1+i for i in range(len(self.seed_edge_pairs))] \n",
    "        self.max_vertex_index = max(self.seed_edge_indices)\n",
    "        \n",
    "        self.seed_edge_zero_skeleton_dict = {max_vertex_index+1+i: edge\n",
    "                                             for i, edge in enumerate(self.seed_edge_pairs)}\n",
    "        \n",
    "        self.vertex_index_dict = {0: self.seed_nodes,\n",
    "                            1: self.seed_edge_indices}\n",
    "        \n",
    "        zero_skeleton_dict = {}\n",
    "        \"\"\"This is the dictionary of simplices. The keys are tuples, with the first entry of the \n",
    "        tuple is the dimension of the simplex, the second entry is the label of the simplex.\n",
    "        The value of this dictionary is the actual position of the simplex in the graph\"\"\" \n",
    "        \n",
    "        for vertex_index in self.vertex_index_dict[0]:\n",
    "            zero_skeleton_dict.update({(0, vertex_index): [vertex_index]})\n",
    "            \n",
    "        for edge_index in self.vertex_index_dict[1]:\n",
    "            zero_skeleton = self.seed_edge_zero_skeleton_dict[edge_index]\n",
    "            zero_skeleton = list(zero_skeleton)\n",
    "            zero_skeleton_dict.update({(1, edge_index): zero_skeleton})\n",
    "        self.zero_skeleton_dict = zero_skeleton_dict\n",
    "    \n",
    "    def levels(self):\n",
    "        levels = [key for key in self.vertex_index_dict]\n",
    "        return levels\n",
    "    \n",
    "    def height(self):\n",
    "        height = max(self.levels())\n",
    "        return height\n",
    "    \n",
    "    def connectivity_update(self):\n",
    "        \"\"\"This is the method that 'connects up' our AdjGraph so that it jumps up from being the adjacency graph\n",
    "        of a k-connected simplicial set to the adjacency graph of a (k+1)-connected simplicial set. This is done\n",
    "        by adding in (k+1)-simplicies where ever our simplicial set contains a non-degenerate boundary of a\n",
    "        standard (k+1)-simplex.\n",
    "        \n",
    "        At a high level, this method procedes as follows:\n",
    "        1. Locate all (0-simplex, k-simplex)-pairs (u, s) such that the vertex u is not contained \n",
    "        in the k-simplex s.\n",
    "        2. For each such pair, let [v_0, v_1, ...., v_k] be the 0-skeleton sk_0(s) of our \n",
    "        k-simplex s.\n",
    "        3. For each resulting (0-simplex, 0-simplex)-pair (u, v_i), for 0<=i<=k, query if \n",
    "        [u, v_i] is a directed edge in our original graph, for all i ([v_i,u]?)\n",
    "        4. If the answer is affirmative for every pair (u, v_i) in Step 3 above, then \n",
    "        [u, v_0, v_1, ...., v_k] is the 0-skeleton of a (k+1)-simplex that needs to be \n",
    "        added to AdjGraph.\n",
    "        \"\"\"\n",
    "        \n",
    "        # record the height of AdjGraph. This becomes the integer k as used in the commentimmediately above:\n",
    "        height = self.height()\n",
    "        \n",
    "        # Create a list of the 0-skeleta of all top-dimensional simplices in our current simpliciat set:\n",
    "        top_dim_zero_skeleta = []\n",
    "        for key in self.zero_skeleton_dict:\n",
    "            if key[0] == height:\n",
    "                top_dim_zero_skeleta.append(self.zero_skeleton_dict[key])\n",
    "        \n",
    "        # Create a list of the 0-skeleta of all 1-simplices, i.e., edges, in our current simplicial set:\n",
    "        edge_zero_skeleta = []\n",
    "        for key in self.zero_skeleton_dict:\n",
    "            if key[0] == 1:\n",
    "                edge_zero_skeleta.append(self.zero_skeleton_dict[key])\n",
    "        \n",
    "        # Begin list of all non-degenerate boundaries of standard (k+1)-simplices \n",
    "        #in our current simplicial set:\n",
    "        top_dim_boundaries = []\n",
    "        # Step 1: Iterate of 0-simplices in our current simplicial set:\n",
    "        for src in self.vertex_index_dict[0]:\n",
    "            # Step 2: iterate over k-simplices in our current simplicial set, extracting the 0-skeleton of each:\n",
    "            for zero_skel in top_dim_zero_skeleta:\n",
    "                # Check that our new 0-simplex doesn't already lie in the 0-skeleton of our k-simplex:\n",
    "                if src in zero_skel:\n",
    "                    forms_boundary_query = False\n",
    "                # If it doesn't, begin Step 3: \n",
    "                else:\n",
    "                    forms_boundary_query = True\n",
    "                    # Check that our original graph contains all necessary edges:\n",
    "                    for dst in zero_skel:\n",
    "                        edge_present_query = ([src, dst] in edge_zero_skeleta)\n",
    "                        forms_boundary_query *= edge_present_query\n",
    "                        if forms_boundary_query == False:\n",
    "                            break\n",
    "                    # If it does, adjoin a new 0-skeleton to our list of (k+1)-simplices to add to AdjGraph:\n",
    "                    if forms_boundary_query == True:\n",
    "                        top_dim_boundaries.append([src]+zero_skel)\n",
    "        \n",
    "        # Update AdjGraph class attributes.\n",
    "        # We've added simplices of one dimension higher, so the height increases:\n",
    "        new_height = 1 + self.height()\n",
    "        \n",
    "        # Since we have new simplices, we have new vertices in AdjGraph, so we need to update vertex indices.\n",
    "        # Fix the old maximum index for vertices in AdjGraph before we update it:\n",
    "        old_max_vertex_index = self.max_vertex_index\n",
    "        # Count how many new simplices we have, so how many new vertices we'll be adding to AdjGraph:\n",
    "        new_simplex_count = len(top_dim_boundaries)\n",
    "        # Update the attribute max_vertex_index by adding the new vertex count to our previous one:\n",
    "        self.max_vertex_index += new_simplex_count\n",
    "        # Create a list of indices for our new simplices, picking up at our previous largest index:\n",
    "        new_indices = [old_max_vertex_index+1+i for i in range(new_simplex_count)]\n",
    "        # Update our dictionary of vertex indices by introducing a new key, corresponding to our new height,\n",
    "        # and define the value at this new key to be the list of all our new indices:\n",
    "        self.vertex_index_dict.update({new_height: new_indices})\n",
    "        # Update our dictionary of 0-skeleta associated to vertices in AdjGraph by introducing one new key\n",
    "        # (simplex dimension, simplex index) for each new simplex in AdjGraph, and define the value at this\n",
    "        # new key to be the 0-skeleton of this simplex:\n",
    "        for i, zero_skeleton in enumerate(top_dim_boundaries):\n",
    "            simplex_index = self.vertex_index_dict[new_height][i]\n",
    "            self.zero_skeleton_dict.update({(new_height, simplex_index): zero_skeleton})\n",
    "    def highest_dimension(self):\n",
    "        for dimension, vertices in self.vertex_index_dict.items():\n",
    "            if dimension == self.height():\n",
    "                if vertices == []:\n",
    "                    return True\n",
    "        return False"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "id": "fd3b0990",
   "metadata": {},
   "outputs": [],
   "source": [
    "\"\"\"Testing implementation with random graphs\"\"\"\n",
    "\n",
    "import timeit\n",
    "import matplotlib.pyplot as plt\n",
    "from time import process_time\n",
    "import random\n",
    "def generate_random_graph(num_nodes, num_edges):\n",
    "    src_edges =[]\n",
    "    dst_edges = []\n",
    "    edges = []\n",
    "    \n",
    "    while len(edges) < num_edges:\n",
    "        src_edge = random.randint(0, num_nodes)\n",
    "        dst_edge = random.randint(0, num_nodes)\n",
    "        if src_edge != dst_edge and (src_edge, dst_edge) not in edges:\n",
    "            src_edges.append(src_edge)\n",
    "            dst_edges.append(dst_edge)\n",
    "            edges.append((src_edge, dst_edge))\n",
    "            \n",
    "    graph = dgl.heterograph({('paper', 'cites', 'paper'): (src_edges, dst_edges)})\n",
    "    return graph, edges"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "487759a4",
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "\r",
      "  0%|                                                                                           | 0/30 [00:00<?, ?it/s]"
     ]
    }
   ],
   "source": [
    "def run_experiment(runs, number_of_nodes):\n",
    "    sparsity_values = list(range(2*number_of_nodes,math.comb(number_of_nodes,2)+1))\n",
    "    \n",
    "    all_sparsity_runs_quotients = []\n",
    "    all_sparsity_runs_dynamic = []\n",
    "    ratio = 0.05\n",
    "\n",
    "    for i in tqdm(range(runs)):\n",
    "        execution_times_quotients = []\n",
    "        execution_times_dynamic = []\n",
    "        for num_edges in sparsity_values:\n",
    "            database_name = 'random{}_{}_{}.db'.format(number_of_nodes, num_edges, i)\n",
    "            file_path = 'random{}_path_{}_{}'.format(number_of_nodes, num_edges, i)\n",
    "            elapsed_time = 0\n",
    "            elapsed_time_dynamic = 0\n",
    "            dgl_G, edges = generate_random_graph(number_of_nodes, num_edges)\n",
    "            random_preprocessing = graph_towers(file_path=file_path,database_name=database_name, graph=dgl_G,ratio=ratio,bottom_level = 5, max_dimension = 20)\n",
    "            if random_preprocessing.maximum_class_size == 1:\n",
    "                execution_times_quotients = execution_times_quotients + [elapsed_time]\n",
    "                pass\n",
    "            else:\n",
    "                start_time = process_time()\n",
    "                random_preprocessing.create_table()\n",
    "                random_preprocessing.create_towers()\n",
    "                random_preprocessing.simplicial_search()\n",
    "                elapsed_time = process_time() - start_time\n",
    "                execution_times_quotients = execution_times_quotients + [elapsed_time]\n",
    "                #os.remove(database_name)\n",
    "            heavy_duty=AdjGraphOriginal(dgl_G)\n",
    "            start_time = process_time()\n",
    "            for index in range(number_of_nodes):\n",
    "                heavy_duty.connectivity_update()\n",
    "                if heavy_duty.highest_dimension() == True:\n",
    "                    break\n",
    "            elapsed_time_dynamic = process_time() - start_time\n",
    "            execution_times_dynamic = execution_times_dynamic + [elapsed_time_dynamic]\n",
    "        all_sparsity_runs_quotients.append(execution_times_quotients)\n",
    "        all_sparsity_runs_dynamic.append(execution_times_dynamic)\n",
    "    all_sparsity_runs_quotients = np.array(all_sparsity_runs_quotients)\n",
    "    all_sparsity_runs_dynamic = np.array(all_sparsity_runs_dynamic)\n",
    "    averages_quotients = np.mean(all_sparsity_runs_quotients, axis=0)\n",
    "    averages_dynamic = np.mean(all_sparsity_runs_dynamic, axis=0)\n",
    "    plt.figure(figsize=(10, 6))\n",
    "    plt.plot(sparsity_values, averages_quotients, \n",
    "             linestyle = 'dotted', label='Graph towers algorithm', color='red',marker='o')\n",
    "    plt.plot(sparsity_values, averages_dynamic, \n",
    "             linestyle = 'dotted', label='Dynamic search', color='blue', marker='x')\n",
    "    plt.xlabel('Number of edges')\n",
    "    plt.ylabel('Average execution time')\n",
    "    plt.title('Average execution time of {} runs of a random graph with {} nodes and ratio {}'.format(runs, number_of_nodes, ratio))\n",
    "    plt.xticks(sparsity_values)\n",
    "    plt.legend()\n",
    "    plt.grid(True)\n",
    "    plt.tight_layout()\n",
    "    plt.show()\n",
    "    plt.savefig('average_execution_time_num_nodes_{}_ratio_{}.png'.format(number_of_nodes,ratio))\n",
    "\n",
    "run_experiment(30,20)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "733b435f",
   "metadata": {},
   "outputs": [],
   "source": [
    "run_experiment(30,25)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ee1b434f",
   "metadata": {},
   "outputs": [],
   "source": [
    "run_experiment(30,15)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5fce5e19",
   "metadata": {},
   "outputs": [],
   "source": [
    "run_experiment(30,10)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ebd2b916",
   "metadata": {},
   "outputs": [],
   "source": [
    "\"\"\"Testing implementation with random graphs\"\"\"\n",
    "def run_experiment2(runs, density):\n",
    "    assert density<=1 and density>=0, \\\n",
    "    'Keyword argument \\\"density\\\" of graph_towers\\'s init method must be between 0 and 1.'\n",
    "    nodes_list = range(2,50)\n",
    "    ratio = 0.1\n",
    "\n",
    "    all_sparsity_runs_quotients = []\n",
    "    all_sparsity_runs_dynamic = []\n",
    "\n",
    "    for i in tqdm(range(runs)):\n",
    "        execution_times_quotients = []\n",
    "        execution_times_dynamic = []\n",
    "        for num_nodes in nodes_list:\n",
    "            num_edges = math.ceil(density*num_nodes*(num_nodes-1))\n",
    "            database_name = 'randomdensity{}_edges{}_nodes{}_run{}.db'.format(density, num_edges, num_nodes, i)\n",
    "            file_path = 'randomdensity{}_edges{}_nodes{}_run{}'.format(density, num_edges, num_nodes, i)\n",
    "            elapsed_time = 0\n",
    "            elapsed_time_dynamic = 0\n",
    "            dgl_G, edges = generate_random_graph(num_nodes, num_edges)\n",
    "            random_preprocessing = graph_towers(file_path=file_path,database_name=database_name, graph=dgl_G,ratio=ratio,bottom_level = 5, max_dimension = 20)\n",
    "            if random_preprocessing.maximum_class_size == 1:\n",
    "                execution_times_quotients = execution_times_quotients + [elapsed_time]\n",
    "                pass\n",
    "            else:\n",
    "                start_time = process_time()\n",
    "                random_preprocessing.create_table()\n",
    "                random_preprocessing.create_towers()\n",
    "                random_preprocessing.simplicial_search()\n",
    "                elapsed_time = process_time() - start_time\n",
    "                execution_times_quotients = execution_times_quotients + [elapsed_time]\n",
    "                #os.remove(database_name)\n",
    "            heavy_duty=AdjGraphOriginal(dgl_G)\n",
    "            start_time = process_time()\n",
    "            for index in range(num_nodes):\n",
    "                heavy_duty.connectivity_update()\n",
    "                if heavy_duty.highest_dimension() == True:\n",
    "                    break\n",
    "            elapsed_time_dynamic = process_time() - start_time\n",
    "            execution_times_dynamic = execution_times_dynamic + [elapsed_time_dynamic]\n",
    "        all_sparsity_runs_quotients.append(execution_times_quotients)\n",
    "        all_sparsity_runs_dynamic.append(execution_times_dynamic)\n",
    "    all_sparsity_runs_quotients = np.array(all_sparsity_runs_quotients)\n",
    "    all_sparsity_runs_dynamic = np.array(all_sparsity_runs_dynamic)\n",
    "    averages_quotients = np.mean(all_sparsity_runs_quotients, axis=0)\n",
    "    averages_dynamic = np.mean(all_sparsity_runs_dynamic, axis=0)\n",
    "    plt.figure(figsize=(10, 6))\n",
    "    plt.plot(nodes_list, averages_quotients, \n",
    "             linestyle = 'dotted', label='Graph towers algorithm', color='red',marker='o')\n",
    "    plt.plot(nodes_list, averages_dynamic, \n",
    "             linestyle = 'dotted', label='Dynamic search', color='blue', marker='x')\n",
    "    plt.xlabel('Number of nodes')\n",
    "    plt.ylabel('Average execution time')\n",
    "    plt.title('Average execution time of {} runs of a random graph with density {} and ratio {}'.format(runs, density, ratio))\n",
    "    plt.xticks(nodes_list)\n",
    "    plt.legend()\n",
    "    plt.grid(True)\n",
    "    plt.tight_layout()\n",
    "    plt.show()\n",
    "    plt.savefig('average_execution_time_density{}_maxnodes{}_ratio_{}.png'.format(density,max(nodes_list)),ratio)\n",
    "\n",
    "run_experiment2(30,0.1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "350463a8",
   "metadata": {},
   "outputs": [],
   "source": [
    "run_experiment2(30,0.2)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.5"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
